# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


# TPR parser and tpr support module
# Copyright (c) 2011 Zhuyi Xue
# Released under the  Lesser GNU Public Licence, v2.1+

"""
TPRParser settings
==================

Definition of constants.

The currently read file format versions are defined in
:data:`SUPPORTED_VERSIONS`.

"""

#: Gromacs TPR file format versions that can be read by the TPRParser.
SUPPORTED_VERSIONS = (
    58,
    73,
    83,
    100,
    103,
    110,
    112,
    116,
    119,
    122,
    127,
    129,
    133,
    134,
    137,
)

# Some constants
STRLEN = 4096
BIG_STRLEN = 1048576
DIM = 3
NR_RBDIHS = 6  # <gromacs-5.1-dir>/src/gromacs/topology/idef.h
NR_CBTDIHS = 6  # <gromacs-5.1-dir>/src/gromacs/topology/idef.h
NR_FOURDIHS = 4  # <gromacs-5.1-dir>/src/gromacs/topology/idef.h
egcNR = 10  # include/types/topolog.h
TPX_TAG_RELEASE = "release"  # <gromacs-5.1-dir>/src/gromacs/fileio/tpxio.c
tpx_version = 103  # <gromacs-5.1-dir>/src/gromacs/fileio/tpxio.c
tpx_generation = 27  # <gromacs-5.1-dir>/src/gromacs/fileio/tpxio.c
tpxv_RestrictedBendingAndCombinedAngleTorsionPotentials = 98
tpxv_GenericInternalParameters = 117
tpxv_VSite2FD = 118
tpxv_AddSizeField = 119
tpxv_VSite1 = 121
tpxv_RemoveTholeRfac = 127


#: Function types from ``<gromacs_dir>/include/types/idef.h``
# fmt: off
(
    F_BONDS, F_G96BONDS, F_MORSE, F_CUBICBONDS,
    F_CONNBONDS, F_HARMONIC, F_FENEBONDS, F_TABBONDS,
    F_TABBONDSNC, F_RESTRBONDS, F_ANGLES, F_G96ANGLES, F_RESTRANGLES,
    F_LINEAR_ANGLES, F_CROSS_BOND_BONDS, F_CROSS_BOND_ANGLES, F_UREY_BRADLEY,
    F_QUARTIC_ANGLES, F_TABANGLES, F_PDIHS, F_RBDIHS, F_RESTRDIHS, F_CBTDIHS,
    F_FOURDIHS, F_IDIHS, F_PIDIHS, F_TABDIHS,
    F_CMAP, F_GB12, F_GB13, F_GB14,
    F_GBPOL, F_NPSOLVATION, F_LJ14, F_COUL14,
    F_LJC14_Q, F_LJC_PAIRS_NB, F_LJ, F_BHAM,
    F_LJ_LR, F_BHAM_LR, F_DISPCORR, F_COUL_SR,
    F_COUL_LR, F_RF_EXCL, F_COUL_RECIP, F_LJ_RECIP, F_DPD,
    F_POLARIZATION, F_WATER_POL, F_THOLE_POL, F_ANHARM_POL,
    F_POSRES, F_FBPOSRES, F_DISRES, F_DISRESVIOL, F_ORIRES,
    F_ORIRESDEV, F_ANGRES, F_ANGRESZ, F_DIHRES,
    F_DIHRESVIOL, F_CONSTR, F_CONSTRNC, F_SETTLE, F_VSITE1,
    F_VSITE2, F_VSITE2FD, F_VSITE3, F_VSITE3FD, F_VSITE3FAD,
    F_VSITE3OUT, F_VSITE4FD, F_VSITE4FDN, F_VSITEN,
    F_COM_PULL, F_DENSITYFITTING, F_EQM, F_ENNPOT, F_EPOT, F_EKIN,
    F_ETOT, F_ECONSERVED, F_TEMP, F_VTEMP_NOLONGERUSED,
    F_PDISPCORR, F_PRES, F_DHDL_CON, F_DVDL,
    F_DKDL, F_DVDL_COUL, F_DVDL_VDW, F_DVDL_BONDED,
    F_DVDL_RESTRAINT, F_DVDL_TEMPERATURE, F_NRE
) = list(range(96))
# fmt: on

#: Function types from ``<gromacs_dir>/src/gmxlib/tpxio.c``
# fmt: off
ftupd = [
    (20, F_CUBICBONDS), (20, F_CONNBONDS), (20, F_HARMONIC), (34, F_FENEBONDS),
    (43, F_TABBONDS), (43, F_TABBONDSNC), (70, F_RESTRBONDS),
    (tpxv_RestrictedBendingAndCombinedAngleTorsionPotentials, F_RESTRANGLES),
    (76, F_LINEAR_ANGLES), (30, F_CROSS_BOND_BONDS), (30, F_CROSS_BOND_ANGLES),
    (30, F_UREY_BRADLEY), (34, F_QUARTIC_ANGLES), (43, F_TABANGLES),
    (tpxv_RestrictedBendingAndCombinedAngleTorsionPotentials, F_RESTRDIHS),
    (tpxv_RestrictedBendingAndCombinedAngleTorsionPotentials, F_CBTDIHS),
    (26, F_FOURDIHS), (26, F_PIDIHS), (43, F_TABDIHS), (65, F_CMAP),
    (60, F_GB12), (61, F_GB13), (61, F_GB14), (72, F_GBPOL),
    (72, F_NPSOLVATION), (41, F_LJC14_Q), (41, F_LJC_PAIRS_NB),
    (32, F_BHAM_LR), (32, F_RF_EXCL), (32, F_COUL_RECIP), (93, F_LJ_RECIP),
    (46, F_DPD), (30, F_POLARIZATION), (36, F_THOLE_POL), (90, F_FBPOSRES),
    (22, F_DISRESVIOL), (22, F_ORIRES), (22, F_ORIRESDEV),
    (26, F_DIHRES), (26, F_DIHRESVIOL), (49, F_VSITE4FDN),
    (50, F_VSITEN), (46, F_COM_PULL), (20, F_EQM),
    (46, F_ECONSERVED), (69, F_VTEMP_NOLONGERUSED), (66, F_PDISPCORR),
    (54, F_DHDL_CON), (76, F_ANHARM_POL), (79, F_DVDL_COUL),
    (79, F_DVDL_VDW,), (79, F_DVDL_BONDED,), (79, F_DVDL_RESTRAINT),
    (79, F_DVDL_TEMPERATURE),
    (tpxv_GenericInternalParameters, F_DENSITYFITTING),
    (tpxv_VSite1, F_VSITE1),
    (tpxv_VSite2FD, F_VSITE2FD),
    (137, F_ENNPOT),
]
# fmt: on

#: Interaction types from ``<gromacs_dir>/gmxlib/ifunc.c``
interaction_types = [
    ("BONDS", "Bond", 2),
    ("G96BONDS", "G96Bond", 2),
    ("MORSE", "Morse", 2),
    ("CUBICBONDS", "Cubic Bonds", 2),
    ("CONNBONDS", "Connect Bonds", 2),
    ("HARMONIC", "Harmonic Pot.", 2),
    ("FENEBONDS", "FENE Bonds", 2),
    ("TABBONDS", "Tab. Bonds", 2),
    ("TABBONDSNC", "Tab. Bonds NC", 2),
    ("RESTRAINTPOT", "Restraint Pot.", 2),
    ("ANGLES", "Angle", 3),
    ("G96ANGLES", "G96Angle", 3),
    ("RESTRANGLES", "Restricted Angles", 3),
    ("LINEAR_ANGLES", "Lin. Angle", 3),
    ("CROSS_BOND_BOND", "Bond-Cross", 3),
    ("CROSS_BOND_ANGLE", "BA-Cross", 3),
    ("UREY_BRADLEY", "U-B", 3),
    ("QANGLES", "Quartic Angles", 3),
    ("TABANGLES", "Tab. Angles", 3),
    ("PDIHS", "Proper Dih.", 4),
    ("RBDIHS", "Ryckaert-Bell.", 4),
    ("RESTRDIHS", "Restricted Dih.", 4),
    ("CBTDIHS", "CBT Dih.", 4),
    ("FOURDIHS", "Fourier Dih.", 4),
    ("IDIHS", "Improper Dih.", 4),
    ("PIDIHS", "Improper Dih.", 4),
    ("TABDIHS", "Tab. Dih.", 4),
    ("CMAP", "CMAP Dih.", 5),
    ("GB12", "GB 1-2 Pol.", 2),
    ("GB13", "GB 1-3 Pol.", 2),
    ("GB14", "GB 1-4 Pol.", 2),
    ("GBPOL", "GB Polarization", None),
    ("NPSOLVATION", "Nonpolar Sol.", None),
    ("LJ14", "LJ-14", 2),
    ("COUL14", "Coulomb-14", None),
    ("LJC14_Q", "LJC-14 q", 2),
    ("LJC_NB", "LJC Pairs NB", 2),
    ("LJ_SR", "LJ (SR)", 2),
    ("BHAM", "Buck.ham (SR)", 2),
    ("LJ_LR", "LJ (LR)", None),
    ("BHAM_LR", "Buck.ham (LR)", None),
    ("DISPCORR", "Disper. corr.", None),
    ("COUL_SR", "Coulomb (SR)", None),
    ("COUL_LR", "Coulomb (LR)", None),
    ("RF_EXCL", "RF excl.", None),
    ("COUL_RECIP", "Coul. recip.", None),
    ("LJ_RECIP", "LJ recip.", None),
    ("DPD", "DPD", None),
    ("POLARIZATION", "Polarization", 2),
    ("WATERPOL", "Water Pol.", 5),
    ("THOLE", "Thole Pol.", 4),
    ("ANHARM_POL", "Anharm. Pol.", 2),
    ("POSRES", "Position Rest.", 1),
    ("FBPOSRES", "Flat-bottom posres", 1),
    ("DISRES", "Dis. Rest.", 2),
    ("DISRESVIOL", "D.R.Viol. (nm)", None),
    ("ORIRES", "Orient. Rest.", 2),
    ("ORDEV", "Ori. R. RMSD", None),
    ("ANGRES", "Angle Rest.", 4),
    ("ANGRESZ", "Angle Rest. Z", 2),
    ("DIHRES", "Dih. Rest.", 4),
    ("DIHRESVIOL", "Dih. Rest. Viol.", None),
    ("CONSTR", "Constraint", 2),
    ("CONSTRNC", "Constr. No Conn.", 2),
    ("SETTLE", "Settle", 3),
    ("VSITE1", "Virtual site 1", 2),
    ("VSITE2", "Virtual site 2", 3),
    ("VSITE2FD", "Virtual site 2fd", 3),
    ("VSITE3", "Virtual site 3", 4),
    ("VSITE3FD", "Virtual site 3fd", 4),
    ("VSITE3FAD", "Virtual site 3fad", 4),
    ("VSITE3OUT", "Virtual site 3out", 4),
    ("VSITE4FD", "Virtual site 4fd", 5),
    ("VSITE4FDN", "Virtual site 4fdn", 5),
    ("VSITEN", "Virtual site N", 2),
    ("COM_PULL", "COM Pull En.", None),
    ("DENSITYFIT", "Density fitting", None),
    ("EQM", "Quantum En.", None),
    ("EPOT", "Potential", None),
    ("EKIN", "Kinetic En.", None),
    ("ETOT", "Total Energy", None),
    ("ECONS", "Conserved En.", None),
    ("TEMP", "Temperature", None),
    ("VTEMP", "Vir. Temp. (not used)", None),
    ("PDISPCORR", "Pres. DC", None),
    ("PRES", "Pressure", None),
    ("DH/DL_CON", "dH/dl constr.", None),
    ("DV/DL", "dVremain/dl", None),
    ("DK/DL", "dEkin/dl", None),
    ("DVC/DL", "dVcoul/dl", None),
    ("DVV/DL", "dVvdw/dl", None),
    ("DVB/DL", "dVbonded/dl", None),
    ("DVR/DL", "dVrestraint/dl", None),
    ("DVT/DL", "dVtemperature/dl", None),
]



import MDAnalysis
import numpy as np

try:
    from MDAnalysis.exceptions import NoDataError
    from MDAnalysisTests.datafiles import DCD, GRO, PSF, TPR, TRZ, XTC, TRZ_psf
except:
    pass


class AtomGroupMethodsBench(object):
    """Benchmarks for the various MDAnalysis
    atomgroup methods.
    """

    # NOTE: the write() method has been
    # excluded as file writing is considered
    # a separate benchmarking category

    params = (10, 100, 1000, 10000)
    param_names = ["num_atoms"]

    def setup(self, num_atoms):
        self.u = MDAnalysis.Universe(GRO)
        self.ag = self.u.atoms[:num_atoms]
        self.weights = np.ones(num_atoms)
        self.vdwradii = {"H": 1.0, "C": 1.0, "N": 1.0, "O": 1.0, "DUMMY": 1.0}
        self.rot_matrix = np.ones((3, 3))
        self.trans = np.ones((4, 4))

    def time_bbox_pbc(self, num_atoms):
        """Benchmark bounding box calculation
        with pbc active.
        """
        self.ag.bbox(pbc=True)

    def time_bbox_no_pbc(self, num_atoms):
        """Benchmark bounding box calculation
        with pbc inactive.
        """
        self.ag.bbox(pbc=False)

    def time_bsphere_pbc(self, num_atoms):
        """Benchmark bounding sphere calculation
        with pbc active.
        """
        self.ag.bsphere(pbc=True)

    def time_bsphere_no_pbc(self, num_atoms):
        """Benchmark bounding sphere calculation
        with pbc inactive.
        """
        self.ag.bsphere(pbc=False)

    def time_center_pbc(self, num_atoms):
        """Benchmark center calculation with
        pbc active.
        """
        self.ag.center(weights=self.weights, pbc=True)

    def time_center_no_pbc(self, num_atoms):
        """Benchmark center calculation with
        pbc inactive.
        """
        self.ag.center(weights=self.weights, pbc=False)

    def time_centroid_pbc(self, num_atoms):
        """Benchmark centroid calculation with
        pbc active.
        """
        self.ag.centroid(pbc=True)

    def time_centroid_no_pbc(self, num_atoms):
        """Benchmark centroid calculation with
        pbc inactive.
        """
        self.ag.centroid(pbc=False)

    def time_concatenate(self, num_atoms):
        """Benchmark atomgroup concatenation."""
        self.ag.concatenate(self.ag)

    def time_difference(self, num_atoms):
        """Benchmark atomgroup difference
        operation.
        """
        self.ag.difference(self.ag)

    def time_groupby(self, num_atoms):
        """Benchmark atomgroup groupby
        operation.
        """
        self.ag.groupby("resnames")

    def time_guess_bonds(self, num_atoms):
        """Benchmark atomgroup bond guessing
        with artificially-seeded vdw values.
        """
        self.ag.guess_bonds(self.vdwradii)

    def time_intersection(self, num_atoms):
        """Benchmark ag intersection."""
        self.ag.intersection(self.ag)

    def time_is_strict_subset(self, num_atoms):
        """Benchmark ag strict subset operation."""
        self.ag.is_strict_subset(self.ag)

    def time_is_strict_superset(self, num_atoms):
        """Benchmark ag strict superset operation."""
        self.ag.is_strict_superset(self.ag)

    def time_isdisjoint(self, num_atoms):
        """Benchmark disjoint operation between
        atomgroups.
        """
        self.ag.isdisjoint(self.ag)

    def time_issubset(self, num_atoms):
        """Benchmark subset operation between
        atomgroups.
        """
        self.ag.issubset(self.ag)

    def time_issuperset(self, num_atoms):
        """Benchmark superset operation between
        atomgroups.
        """
        self.ag.issuperset(self.ag)

    def time_pack_into_box(self, num_atoms):
        """Benchmark shifting atoms of ag
        into primary unit cell, using
        default parameters.
        """
        self.ag.pack_into_box()

    def time_rotate(self, num_atoms):
        """Benchmark simple rotation operation
        on atomgroup.
        """
        self.ag.rotate(self.rot_matrix)

    def time_rotateby(self, num_atoms):
        """Benchmark rotation by an angle
        of the ag coordinates.
        """
        self.ag.rotateby(angle=45, axis=[1, 0, 0])

    def time_split(self, num_atoms):
        """Benchmark ag splitting into
        multiple ags based on a simple
        criterion.
        """
        self.ag.split("residue")

    def time_subtract(self, num_atoms):
        """Benchmark ag subtraction."""
        self.ag.subtract(self.ag)

    def time_symmetric_difference(self, num_atoms):
        """Benchmark ag symmetric difference
        operation.
        """
        self.ag.symmetric_difference(self.ag)

    def time_transform(self, num_atoms):
        """Benchmark application of transformation
        matrix to atomgroup.
        """
        self.ag.transform(self.trans)

    def time_translate(self, num_atoms):
        """Benchmark the application of a
        translation vector to the ag
        coordinates.
        """
        self.ag.translate([0, 0.5, 1])

    def time_union(self, num_atoms):
        """Benchmark union operation
        on atomgroups.
        """
        self.ag.union(self.ag)

    def time_wrap(self, num_atoms):
        """Benchmark wrap() operation on
        atomgroup with default params.
        """
        self.ag.wrap()


class AtomGroupAttrsBench(object):
    """Benchmarks for the various MDAnalysis
    atomgroup attributes.
    """

    params = (10, 100, 1000, 10000)
    param_names = ["num_atoms"]

    def setup(self, num_atoms):
        self.u = MDAnalysis.Universe(GRO)
        self.ag = self.u.atoms[:num_atoms]

    def time_angle(self, num_atoms):
        """Benchmark simple angle
        calculation. Requires ag
        with three atoms.
        """
        self.ag[:3].angle

    def time_atoms(self, num_atoms):
        """Benchmark returning of identical
        atomgroup.
        """
        self.ag.atoms

    def time_dihedral(self, num_atoms):
        """Benchmark Dihedral object
        creation time. Requires ag of
        size 4.
        """
        self.ag[:4].dihedral

    # TODO: use universe / ag that
    # is suitable for force calc
    def time_forces(self, num_atoms):
        """Benchmark atomgroup force
        calculation.
        """
        try:
            self.ag.forces
        except NoDataError:
            pass

    # TODO: use universe / ag that
    # is suitable for velocity extraction
    def time_velocity(self, num_atoms):
        """Benchmark atomgroup velocity
        values return.
        """
        try:
            self.ag.velocities
        except NoDataError:
            pass

    def time_improper(self, num_atoms):
        """Benchmark improper dihedral
        calculation. Requires ag of size
        4.
        """
        self.ag[:4].improper

    def time_indices(self, num_atoms):
        """Benchmark atom index calculation."""
        self.ag.ix

    def time_atomcount(self, num_atoms):
        """Benchmark counting of atoms in
        atomgroup.
        """
        self.ag.n_atoms

    def time_residuecount(self, num_atoms):
        """Benchmark counting of residues in
        atomgroup.
        """
        self.ag.n_residues

    def time_segmentcount(self, num_atoms):
        """Benchmark counting of segments in
        atomgroup.
        """
        self.ag.n_segments

    def time_positions(self, num_atoms):
        """Benchmark returning the positions
        of the atoms in the group.
        """
        self.ag.positions

    def time_residues(self, num_atoms):
        """Benchmark creation of the ResidueGroup
        from the AtomGroup.
        """
        self.ag.residues

    def time_segments(self, num_atoms):
        """Benchmark determination of sorted
        SegmentGroup from AtomGroup.
        """
        self.ag.segments

    def time_ts(self, num_atoms):
        """Benchmark returning of a timestep
        instance from atomgroup.
        """
        self.ag.ts

    def time_unique(self, num_atoms):
        """Benchmark determination of unique
        elements in atomgroup.
        """
        self.ag.unique

    def time_bond(self, num_atoms):
        """Benchmark Bond object creation.
        Requires ag of size 2.
        """
        self.ag[:2].bond


class CompoundSplitting(object):
    """Test how fast can we split compounds into masks and apply them

    The benchmark used in Issue #3000. Parameterizes multiple compound number
    and size combinations.
    """

    params = [
        (100, 10000, 1000000),  # n_atoms
        (1, 10, 100),  # n_compounds
        (True, False),  # homogeneous
        (True, False),
    ]  # contiguous

    def setup(self, n_atoms, n_compounds, homogeneous, contiguous):
        rg = np.random.Generator(np.random.MT19937(3000))

        # some parameter screening for nonsensical combinations.
        if n_compounds > n_atoms:
            raise NotImplementedError

        if n_compounds == 1 and not (homogeneous and contiguous):
            raise NotImplementedError

        if n_compounds == n_atoms:
            if not (homogeneous and contiguous):
                raise NotImplementedError
            compound_indices = np.arange(n_compounds)
        elif homogeneous:
            ats_per_compound, remainder = divmod(n_atoms, n_compounds)
            if remainder:
                raise NotImplementedError
            compound_indices = np.tile(
                np.arange(n_compounds), (ats_per_compound, 1)
            ).T.ravel()
        else:
            compound_indices = np.sort(
                np.floor(rg.random(n_atoms) * n_compounds).astype(np.int)
            )

        unique_indices = np.unique(compound_indices)
        if len(unique_indices) != n_compounds:
            raise RuntimeError

        if not contiguous:
            rg.shuffle(compound_indices)

        self.u = MDAnalysis.Universe.empty(
            n_atoms,
            n_residues=n_compounds,
            n_segments=1,
            atom_resindex=compound_indices,
            trajectory=True,
        )
        self.u.atoms.positions = rg.random((n_atoms, 3), dtype=np.float32) * 100
        self.u.dimensions = [50, 50, 50, 90, 90, 90]

    def time_center_compounds(self, *args):
        self.u.atoms.center(None, compound="residues")


class FragmentFinding(object):
    """Test how quickly we find fragments (distinct molecules from bonds)"""

    # if we try to parametrize over topology &
    # trajectory formats asv will use all
    # possible combinations, so instead handle
    # this in setup()
    params = (
        "large_fragment_small_solvents",
        "large_fragment",
        "polymer_chains",  # 20ish polymer chains
    )
    param_names = ["universe_type"]

    def setup(self, universe_type):
        if universe_type == "large_fragment_small_solvents":
            univ = (TPR, XTC)
        elif universe_type == "large_fragment":
            univ = (PSF, DCD)
        else:
            univ = (TRZ_psf, TRZ)
        self.u = MDAnalysis.Universe(*univ)

    def time_find_fragments(self, universe_type):
        frags = self.u.atoms.fragments


class FragmentCaching(FragmentFinding):
    """Test how quickly we find cached fragments"""

    def setup(self, universe_type):
        super(FragmentCaching, self).setup(universe_type)
        frags = self.u.atoms.fragments  # Priming the cache

    def time_find_cached_fragments(self, universe_type):
        frags = self.u.atoms.fragments



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
FHI-AIMS file format --- :mod:`MDAnalysis.coordinates.FHIAIMS`
==============================================================

Classes to read and write `FHI-AIMS`_ coordinate files.

The cell vectors are specified by the (optional) lines with the
``lattice_vector`` tag::

    lattice_vector x  y  z

where x, y, and z are expressed in ångström (Å).

.. Note::

   In the original `FHI-AIMS format`_, up to three lines with
   ``lattice_vector`` are allowed (order matters) where the absent line implies
   no periodicity in that direction.  In MDAnalysis, only the case of no
   ``lattice_vector`` or three ``lattice_vector`` lines are allowed.

Atomic positions and names are specified either by the ``atom`` or by the
``atom_frac`` tags::

    atom           x  y  z  name
    atom_frac      nx ny nz name

where x, y, and z are expressed in ångström, and nx, ny and nz are real numbers
in [0, 1] and are used to compute the atomic positions in units of the basic
cell.

Atomic velocities can be added on the line right after the corresponding
``atom`` in units of Å/ps using the ``velocity`` tag::

    velocity      vx vy vz


The field name is a string identifying the atomic species.  See also the
specifications in the official `FHI-AIMS format`_.

Classes
-------

.. autoclass:: FHIAIMSReader
   :members:
   :inherited-members:

.. autoclass:: FHIAIMSWriter
   :members:
   :inherited-members:

Developer notes: ``FHIAIMSWriter`` format strings
-------------------------------------------------

The :class:`FHIAIMSWriter` class has a :attr:`FHIAIMSWriter.fmt`
attribute, which is a dictionary of different strings for writing
lines in ``.in`` files.  These are as follows:

``xyz``
  An atom line without velocities.  Requires that the `name` and
  `pos` keys be supplied.  E.g.::

     fmt['xyz'].format(pos=(0.0, 1.0, 2.0), name='O')

``vel``
  An line that specifies velocities::

     fmt['xyz'].format(vel=(0.1, 0.2, 0.3))

``box_triclinic``
  The (optional) initial lines of the file which gives box dimensions.
  Requires the `box` keyword, as a length 9 vector.  This is a flattened
  version of the (3, 3) triclinic vector representation of the unit
  cell.


.. Links

.. _FHI-AIMS: https://aimsclub.fhi-berlin.mpg.de/
.. _FHI-AIMS format: https://doi.org/10.6084/m9.figshare.12413477.v1

"""
import re

import itertools
import warnings

import numpy as np

from . import base
from .core import triclinic_box, triclinic_vectors
from ..exceptions import NoDataError
from ..lib import util
from ..lib import mdamath


class FHIAIMSReader(base.SingleFrameReaderBase):
    """Reader for the FHIAIMS geometry format.

    Single frame reader for the `FHI-AIMS`_ input file format.  Reads
    geometry (3D and molecules only), positions (absolut or fractional),
    velocities if given, all according to the `FHI-AIMS format`_
    specifications

    """

    format = ["IN", "FHIAIMS"]
    units = {"time": "ps", "length": "Angstrom", "velocity": "Angstrom/ps"}

    def _read_first_frame(self):
        with util.openany(self.filename, "rt") as fhiaimsfile:
            relative, positions, velocities, lattice_vectors = [], [], [], []
            skip_tags = ["#", "initial_moment"]
            oldline = ""
            for line in fhiaimsfile:
                line = line.strip()
                if line.startswith("atom"):
                    positions.append(line.split()[1:-1])
                    relative.append("atom_frac" in line)
                    oldline = line
                    continue
                if line.startswith("velocity"):
                    if not "atom" in oldline:
                        raise ValueError(
                            "Non-conforming line (velocity must follow atom): ({0})in FHI-AIMS input file {0}".format(
                                line, self.filename
                            )
                        )
                    velocities.append(line.split()[1:])
                    oldline = line
                    continue
                if line.startswith("lattice"):
                    lattice_vectors.append(line.split()[1:])
                    oldline = line
                    continue
                if any([line.startswith(tag) for tag in skip_tags]):
                    oldline = line
                    continue
                raise ValueError(
                    "Non-conforming line: ({0})in FHI-AIMS input file {0}".format(
                        line, self.filename
                    )
                )

        # positions and velocities are lists of lists of strings; they will be
        # cast to np.arrays(..., dtype=float32) during assignment to ts.positions/ts.velocities
        lattice_vectors = np.asarray(lattice_vectors, dtype=np.float32)

        if len(velocities) not in (0, len(positions)):
            raise ValueError(
                "Found incorrect number of velocity tags ({0}) in the FHI-AIMS file, should be {1}.".format(
                    len(velocities), len(positions)
                )
            )
        if len(lattice_vectors) not in (0, 3):
            raise ValueError(
                "Found partial periodicity in FHI-AIMS file. This cannot be handled by MDAnalysis."
            )
        if len(lattice_vectors) == 0 and any(relative):
            raise ValueError(
                "Found relative coordinates in FHI-AIMS file without lattice info."
            )

        # create Timestep

        self.n_atoms = n_atoms = len(positions)
        self.ts = ts = self._Timestep(n_atoms, **self._ts_kwargs)
        ts.positions = positions

        if len(lattice_vectors) > 0:
            ts.dimensions = triclinic_box(*lattice_vectors)
            ts.positions[relative] = np.matmul(
                ts.positions[relative], lattice_vectors
            )

        if len(velocities) > 0:
            ts.velocities = velocities

        self.ts.frame = 0  # 0-based frame number

    def Writer(self, filename, n_atoms=None, **kwargs):
        """Returns a FHIAIMSWriter for *filename*.

        Parameters
        ----------
        filename: str
            filename of the output FHI-AIMS file

        Returns
        -------
        :class:`FHIAIMSWriter`

        """
        if n_atoms is None:
            n_atoms = self.n_atoms
        return FHIAIMSWriter(filename, n_atoms=n_atoms, **kwargs)


class FHIAIMSWriter(base.WriterBase):
    """FHI-AIMS Writer.

    Single frame writer for the `FHI-AIMS`_ format.  Writes geometry (3D and
    molecules only), positions (absolut only), velocities if given, all
    according to the `FHI-AIMS format`_ specifications.

    If no atom names are given, it will set each atom name to "X".

    """

    format = ["IN", "FHIAIMS"]
    units = {"time": None, "length": "Angstrom", "velocity": "Angstrom/ps"}

    #: format strings for the FHI-AIMS file (all include newline)
    fmt = {
        # coordinates output format, see  https://doi.org/10.6084/m9.figshare.12413477.v1
        "xyz": "atom {pos[0]:12.8f} {pos[1]:12.8f} {pos[2]:12.8f} {name:<3s}\n",
        "vel": "velocity {vel[0]:12.8f} {vel[1]:12.8f} {vel[2]:12.8f}\n",
        # unitcell
        "box_triclinic": "lattice_vector {box[0]:12.8f} {box[1]:12.8f} {box[2]:12.8f}\nlattice_vector {box[3]:12.8f} {box[4]:12.8f} {box[5]:12.8f}\nlattice_vector {box[6]:12.8f} {box[7]:12.8f} {box[8]:12.8f}\n",
    }

    def __init__(self, filename, convert_units=True, n_atoms=None, **kwargs):
        """Set up the FHI-AIMS Writer

        Parameters
        -----------
        filename : str
            output filename
        n_atoms : int (optional)
            number of atoms

        """
        self.filename = util.filename(filename, ext=".in", keep=True)
        self.n_atoms = n_atoms

    def _write_next_frame(self, obj):
        """Write selection at current trajectory frame to file.

        Parameters
        -----------
        obj : AtomGroup or Universe


        .. versionchanged:: 2.0.0
           Support for deprecated Timestep argument has now been removed.
           Use AtomGroup or Universe as an input instead.
        """
        # write() method that complies with the Trajectory API
        try:
            # make sure to use atoms (Issue 46)
            ag = obj.atoms
            # can write from selection == Universe (Issue 49)
        except AttributeError:
            errmsg = "Input obj is neither an AtomGroup or Universe"
            raise TypeError(errmsg) from None

        # Check for topology information
        missing_topology = []
        try:
            names = ag.names
        except (AttributeError, NoDataError):
            names = itertools.cycle(("X",))
            missing_topology.append("names")

        try:
            atom_indices = ag.ids
        except (AttributeError, NoDataError):
            atom_indices = range(1, ag.n_atoms + 1)
            missing_topology.append("ids")

        if missing_topology:
            warnings.warn(
                "Supplied AtomGroup was missing the following attributes: "
                "{miss}. These will be written with default values. "
                "".format(miss=", ".join(missing_topology))
            )

        positions = ag.positions
        try:
            velocities = ag.velocities
            has_velocities = True
        except (AttributeError, NoDataError):
            has_velocities = False

        with util.openany(self.filename, "wt") as output_fhiaims:
            # Lattice
            try:  # for AtomGroup/Universe
                tri_dims = obj.universe.trajectory.ts.triclinic_dimensions
            except AttributeError:  # for Timestep
                tri_dims = obj.triclinic_dimensions
            # full output
            if tri_dims is not None:
                output_fhiaims.write(
                    self.fmt["box_triclinic"].format(box=tri_dims.flatten())
                )

            # Atom descriptions and coords
            # Dont use enumerate here,
            # all attributes could be infinite cycles!
            for atom_index, name in zip(range(ag.n_atoms), names):
                output_fhiaims.write(
                    self.fmt["xyz"].format(
                        pos=positions[atom_index], name=name
                    )
                )
                if has_velocities:
                    output_fhiaims.write(
                        self.fmt["vel"].format(vel=velocities[atom_index])
                    )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import assert_equal, assert_almost_equal

import MDAnalysis as mda
from MDAnalysis.exceptions import DuplicateWarning, NoDataError
from MDAnalysisTests.datafiles import PSF, DCD, GRO, PDB_multipole
from MDAnalysisTests.core.util import UnWrapUniverse
import pytest

levels = ("atoms", "residues", "segments")


class TestAccumulate(object):
    """Tests the functionality of *Group.accumulate()."""

    @pytest.fixture(params=levels)
    def group(self, request):
        u = mda.Universe(PSF, DCD)
        return getattr(u, request.param)

    def test_accumulate_str_attribute(self, group):
        assert_almost_equal(
            group.accumulate("masses"), np.sum(group.atoms.masses)
        )

    def test_accumulate_different_func(self, group):
        assert_almost_equal(
            group.accumulate("masses", function=np.prod),
            np.prod(group.atoms.masses),
        )

    @pytest.mark.parametrize(
        "name, compound",
        (
            ("resindices", "residues"),
            ("segindices", "segments"),
            ("molnums", "molecules"),
            ("fragindices", "fragments"),
        ),
    )
    @pytest.mark.parametrize("level", levels)
    def test_accumulate_str_attribute_compounds(self, name, compound, level):
        u = UnWrapUniverse()
        group = getattr(u, level)
        ref = [sum(a.masses) for a in group.atoms.groupby(name).values()]
        vals = group.accumulate("masses", compound=compound)
        assert_almost_equal(vals, ref, decimal=5)

    def test_accumulate_wrongname(self, group):
        with pytest.raises(AttributeError):
            group.accumulate("foo")

    def test_accumulate_wrongcomponent(self, group):
        with pytest.raises(ValueError):
            group.accumulate("masses", compound="foo")

    @pytest.mark.parametrize("level", levels)
    def test_accumulate_nobonds(self, level):
        group = getattr(mda.Universe(GRO), level)
        with pytest.raises(NoDataError):
            group.accumulate("masses", compound="fragments")

    @pytest.mark.parametrize("level", levels)
    def test_accumulate_nomolnums(self, level):
        group = getattr(mda.Universe(GRO), level)
        with pytest.raises(NoDataError):
            group.accumulate("masses", compound="molecules")

    def test_accumulate_array_attribute(self, group):
        a = np.ones((len(group.atoms), 2, 5))
        assert_equal(group.accumulate(a), np.sum(a, axis=0))

    def test_accumulate_array_attribute_wrongshape(self, group):
        with pytest.raises(ValueError):
            group.accumulate(np.ones(len(group.atoms) - 1))

    @pytest.mark.parametrize(
        "name, compound",
        (
            ("resindices", "residues"),
            ("segindices", "segments"),
            ("molnums", "molecules"),
            ("fragindices", "fragments"),
        ),
    )
    @pytest.mark.parametrize("level", levels)
    def test_accumulate_array_attribute_compounds(self, name, compound, level):
        u = UnWrapUniverse()
        group = getattr(u, level)
        ref = [
            np.ones((len(a), 2, 5)).sum(axis=0)
            for a in group.atoms.groupby(name).values()
        ]
        assert_equal(
            group.accumulate(
                np.ones((len(group.atoms), 2, 5)), compound=compound
            ),
            ref,
        )


class TestTotals(object):
    """Tests the functionality of *Group.total*() like total_mass
    and total_charge.
    """

    @pytest.fixture(params=levels)
    def group(self, request):
        u = mda.Universe(PSF, DCD)
        return getattr(u, request.param)

    def test_total_charge(self, group):
        assert_almost_equal(group.total_charge(), -4.0, decimal=4)

    @pytest.mark.parametrize(
        "name, compound",
        (
            ("resids", "residues"),
            ("segids", "segments"),
            ("fragindices", "fragments"),
        ),
    )
    def test_total_charge_compounds(self, group, name, compound):
        ref = [sum(a.charges) for a in group.atoms.groupby(name).values()]
        assert_almost_equal(group.total_charge(compound=compound), ref)

    @pytest.mark.filterwarnings(  # Prevents regression of issue #2990
        "error:"
        "Using a non-tuple sequence for multidimensional indexing is deprecated:"
        "FutureWarning"
    )
    def test_total_charge_duplicates(self, group):
        group2 = group + group[0]
        ref = group.total_charge() + group[0].charge
        with pytest.warns(DuplicateWarning) as w:
            assert_almost_equal(group2.total_charge(), ref)
            assert len(w) == 1

    def test_total_mass(self, group):
        assert_almost_equal(group.total_mass(), 23582.043)

    @pytest.mark.parametrize(
        "name, compound",
        (
            ("resids", "residues"),
            ("segids", "segments"),
            ("fragindices", "fragments"),
        ),
    )
    def test_total_mass_compounds(self, group, name, compound):
        ref = [sum(a.masses) for a in group.atoms.groupby(name).values()]
        assert_almost_equal(group.total_mass(compound=compound), ref)

    @pytest.mark.filterwarnings(  # Prevents regression of issue #2990
        "error:"
        "Using a non-tuple sequence for multidimensional indexing is deprecated:"
        "FutureWarning"
    )
    def test_total_mass_duplicates(self, group):
        group2 = group + group[0]
        ref = group.total_mass() + group2[0].mass
        with pytest.warns(DuplicateWarning) as w:
            assert_almost_equal(group2.total_mass(), ref)
            assert len(w) == 1


class TestMultipole(object):
    """Tests the functionality of *Group.*_moment() like dipole_moment
    and quadrupole_moment.
    """

    @pytest.fixture(scope="class")
    def u(self):
        u = mda.Universe(PDB_multipole)
        u.add_TopologyAttr(
            "charges",
            [
                -0.076,
                0.019,
                0.019,
                0.019,
                0.019,  # methane [:5]
                -0.003,
                0.001,
                0.001,
                0.001,  # ammonia [5:9]
                -0.216,
                0.108,
                0.108,  # water [9:12]
                -0.25,
                0.034,
                0.35,
                0.037,
                -0.25,
                0.034,
                0.034,
            ],
        )  # acetate [12:]
        lx, ly, lz = np.max(u.atoms.positions, axis=0) - np.min(
            u.atoms.positions, axis=0
        )
        u.dimensions = np.array([lx, ly, lz, 90, 90, 90], dtype=float)

        return u

    @pytest.fixture(scope="class")
    def group(self, u):
        group = u.select_atoms("all")
        return group

    @pytest.fixture(scope="class")
    def methane(self, u):
        group = u.select_atoms("resname CH4")
        return group

    # Dipole
    def test_dipole_moment_com(self, methane):
        dipoles = [
            methane.dipole_moment(unwrap=True),
            methane.dipole_moment(wrap=True),
            methane.dipole_moment(),
        ]
        assert_almost_equal(dipoles, [0.0, 0.2493469, 0.0])

    def test_dipole_moment_no_center(self, group):
        try:
            group.dipole_moment(unwrap=True, center="not supported")
        except ValueError as e:
            assert "not supported" in e.args[0]

    def test_dipole_moment_residues_com_coc(self, group):
        compound = "residues"
        (_, _, n_compounds) = group.atoms._split_by_compound_indices(compound)
        dipoles_com = group.dipole_moment(compound=compound, unwrap=False)
        dipoles_coc = group.dipole_moment(
            compound=compound, unwrap=False, center="charge"
        )

        assert_almost_equal(
            dipoles_com, np.array([0.0, 0.0010198, 0.1209898, 0.5681058])
        )
        assert_almost_equal(dipoles_com[:3], dipoles_coc[:3])
        assert dipoles_com[3] != dipoles_coc[3]
        assert len(dipoles_com) == n_compounds

    def test_dipole_moment_segment(self, methane):
        compound = "segments"
        (_, _, n_compounds) = methane.atoms._split_by_compound_indices(
            compound
        )
        dipoles = methane.dipole_moment(compound=compound, unwrap=True)
        assert_almost_equal(dipoles, [0.0]) and len(dipoles) == n_compounds

    def test_dipole_moment_fragments(self, group):
        compound = "fragments"
        (_, _, n_compounds) = group.atoms._split_by_compound_indices(compound)
        dipoles = group.dipole_moment(compound=compound, unwrap=False)
        assert_almost_equal(
            dipoles, np.array([0.0, 0.0010198, 0.1209898, 0.5681058])
        ) and len(dipoles) == n_compounds

    # Quadrupole
    def test_quadrupole_moment_com(self, methane):
        quadrupoles = [
            methane.quadrupole_moment(unwrap=True),
            methane.quadrupole_moment(wrap=True),
            methane.quadrupole_moment(),
        ]
        assert_almost_equal(quadrupoles, [0.0, 0.4657596, 0.0])

    def test_quadrupole_moment_coc(self, group):
        assert_almost_equal(
            group.quadrupole_moment(unwrap=False, center="charge"),
            0.9769951421535777,
        )

    def test_quadrupole_moment_no_center(self, group):
        try:
            group.quadrupole_moment(unwrap=True, center="not supported")
        except ValueError as e:
            assert "not supported" in e.args[0]

    def test_quadrupole_moment_residues(self, group):
        compound = "residues"
        (_, _, n_compounds) = group.atoms._split_by_compound_indices(compound)

        quadrupoles = group.quadrupole_moment(compound=compound, unwrap=False)
        assert_almost_equal(
            quadrupoles, np.array([0.0, 0.0011629, 0.1182701, 0.6891748])
        ) and len(quadrupoles) == n_compounds

    def test_quadrupole_moment_segment(self, methane):
        compound = "segments"
        (_, _, n_compounds) = methane.atoms._split_by_compound_indices(
            compound
        )
        quadrupoles = methane.quadrupole_moment(compound=compound, unwrap=True)
        assert_almost_equal(quadrupoles, [0.0]) and len(
            quadrupoles
        ) == n_compounds

    def test_quadrupole_moment_fragments(self, group):
        compound = "fragments"
        (_, _, n_compounds) = group.atoms._split_by_compound_indices(compound)

        quadrupoles = group.quadrupole_moment(compound=compound, unwrap=False)
        assert_almost_equal(
            quadrupoles, np.array([0.0, 0.0011629, 0.1182701, 0.6891748])
        ) and len(quadrupoles) == n_compounds



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda

from MDAnalysisTests.topology.base import ParserBase

from MDAnalysisTests.datafiles import (
    PDBQT_input,  # pdbqt_inputpdbqt.pdbqt
    PDBQT_tyrosol,  # tyrosol.pdbqt.bz2
)

from numpy.testing import assert_allclose


class TestPDBQT(ParserBase):
    parser = mda.topology.PDBQTParser.PDBQTParser
    ref_filename = PDBQT_input
    expected_attrs = [
        "ids",
        "names",
        "charges",
        "types",
        "altLocs",
        "resids",
        "resnames",
        "segids",
        "chainIDs",
        "record_types",
        "icodes",
        "occupancies",
        "tempfactors",
    ]

    guessed_attrs = ["masses"]
    expected_n_atoms = 1805
    expected_n_residues = 199  # resids go 2-102 then 2-99
    expected_n_segments = 2  # res2-102 are A, 2-99 are B

    def test_guessed_masses(self, filename):
        u = mda.Universe(filename)
        assert_allclose(
            u.atoms.masses[:7], [14.007, 0.0, 0.0, 12.011, 12.011, 0.0, 12.011]
        )


def test_footnote():
    """just test that the Universe is built even in the presence of a
    footnote before the charges.
    """
    mda.Universe(PDBQT_tyrosol)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Generation and Analysis of X3DNA helicoidal parameter profiles --- :mod:`MDAnalysis.analysis.legacy.x3dna`
==========================================================================================================

:Author: Elizabeth Denning
:Year: 2013-2014
:Copyright: Lesser GNU Public License v2.1+

.. versionadded:: 0.8
.. versionchanged:: 0.16.0
   This module is difficult to test due to restrictions on the X3DNA_ code. It
   is therefore considered unmaintained and legacy code. It was moved to the
   :mod:`MDAnalysis.analysis.legacy` package (see `issue 906`_)

.. _`issue 906`: https://github.com/MDAnalysis/mdanalysis/issues/906

With the help of this module, X3DNA_ can be run on frames in a trajectory. Data
can be combined and analyzed. X3DNA_ :footcite:p:`Lu2003,Lu2008` must be installed
separately.


.. rubric:: References

.. footbibliography::

.. _X3DNA: http://x3dna.org/


Example applications
--------------------

Single structure
~~~~~~~~~~~~~~~~

B-DNA structure::

   from MDAnalysis.analysis.x3dna import X3DNA, X3DNAtraj
   from MDAnalysis.tests.datafiles import PDB_X3DNA

   # set path to your x3dna binary in bashrc file
   H = X3DNA(PDB_X3DNA, executable="x3dna_ensemble analyze -b 355d.bps -p pdbfile")
   H.run()
   H.collect()
   H.plot()


Trajectory
~~~~~~~~~~

Analyzing a trajectory::

  u = MDAnalysis.Universe(psf, trajectory)
  H = X3DNAtraj(u, ...)
  H.run()
  H.plot()
  H.save()

The profiles are available as the attribute :attr:`X3DNAtraj.profiles`
(``H.profiles`` in the example) and are indexed by frame number but
can also be indexed by an arbitrary order parameter as shown in the
next example.



Analysis classes
----------------

.. autoclass:: X3DNA
   :members:
   :inherited-members:

   .. attribute:: profiles

      ``x3dna_ensemble analyze -b 355d.bps -p pdbfile attribute``:
      After running :meth:`X3DNA.collect`, this dict contains all the
      X3DNA profiles, indexed by the frame number. If only a single
      frame was analyzed then this will be ``X3DNA.profiles[0]``. Note
      that the order is random; one needs to sort the keys first.

.. autoclass:: X3DNAtraj
   :members:
   :inherited-members:

   .. attribute:: profiles

      After running :meth:`X3DNA.collect`, this dict contains all the
      X3DNA profiles, indexed by the frame number.



Utilities
---------

.. autoexception:: ApplicationError

"""
import errno
import glob
import logging
import os
import os.path
import shutil
import subprocess
import tempfile
import textwrap
import warnings
from collections import OrderedDict

import matplotlib.pyplot as plt
import numpy as np

from MDAnalysis import ApplicationError
from MDAnalysis.lib.util import asiterable, deprecate, realpath, which

logger = logging.getLogger("MDAnalysis.analysis.x3dna")


@deprecate(
    release="2.7.0",
    remove="3.0.0",
    message=(
        "X3DNA module is deprecated and will be removed in"
        "MDAnalysis 3.0.0, see #3788"
    ),
)
def mean_std_from_x3dnaPickle(profile):
    """Get mean and standard deviation of helicoidal parameters from a saved `profile`.

    The `profile` should have been saved with :meth:`BaseX3DNA.save`. Then
    load it with ::

      profile = cPickle.load(open("x3dna.pickle"))
      h_mean, h_std = mean_std_from_x3dnaPickle(profile)

    Arguments
    ---------
    profile : dict
        A :attr:`X3DNA.profiles` dict with results from the
        :class:`X3DNA` analysis.

    Returns
    -------
    (list, list)
        The tuple contains two lists with the means and the standard deviations
        for the helicoidal parameters. The order for both lists is ``[shear,
        stretch, stagger, buckle, propeller, opening, shift, slide, rise, tilt,
        roll, twist]``.

    .. deprecated:: 2.7.0
        X3DNA will be removed in 3.0.0.
    """
    warnings.warn(
        "X3DNA module is deprecated and will be removed in MDAnalysis 3.0, see #3788",
        category=DeprecationWarning,
    )
    if profile.x3dna_param is False:
        # fmt: off
        bp_shear, bp_stretch, bp_stagger, bp_rise, bp_shift, bp_slide, bp_buckle, bp_prop, bp_open, bp_tilt, bp_roll,\
            bp_twist = [], [], [], [], [], [], [], [], [], [], [], []
        # fmt: on
        for i in range(len(profile)):
            bp_shear.append(profile.values()[i].Shear)
            bp_stretch.append(profile.values()[i].Stretch)
            bp_stagger.append(profile.values()[i].Stagger)
            bp_buckle.append(profile.values()[i].Buckle)
            bp_prop.append(profile.values()[i].Propeller)
            bp_open.append(profile.values()[i].Opening)
            bp_rise.append(profile.values()[i].Rise)
            bp_shift.append(profile.values()[i].Shift)
            bp_slide.append(profile.values()[i].Slide)
            bp_tilt.append(profile.values()[i].Tilt)
            bp_roll.append(profile.values()[i].Roll)
            bp_twist.append(profile.values()[i].Twist)
        # fmt: off
        bp_shear, bp_stretch, bp_stagger, bp_rise, bp_shift, bp_slide, bp_buckle, bp_prop, bp_open, bp_tilt, bp_roll,\
            bp_twist = np.array(bp_shear), np.array(bp_stretch), np.array(bp_stagger), np.array(bp_rise),\
            np.array(bp_shift), np.array(bp_slide), np.array(bp_buckle), np.array(bp_prop), \
            np.array(bp_open), np.array(bp_tilt), np.array(bp_roll), np.array(bp_twist)
        # fmt: on
        na_avg, na_std = [], []
        for j in range(len(bp_shear[0])):
            na_avg.append(
                [
                    np.mean(bp_shear[:, j]),
                    np.mean(bp_stretch[:, j]),
                    np.mean(bp_stagger[:, j]),
                    np.mean(bp_buckle[:, j]),
                    np.mean(bp_prop[:, j]),
                    np.mean(bp_open[:, j]),
                    np.mean(bp_shift[:, j]),
                    np.mean(bp_slide[:, j]),
                    np.mean(bp_rise[:, j]),
                    np.mean(bp_tilt[:, j]),
                    np.mean(bp_roll[:, j]),
                    np.mean(bp_twist[:, j]),
                ]
            )
            na_std.append(
                [
                    np.std(bp_shear[:, j]),
                    np.std(bp_stretch[:, j]),
                    np.std(bp_stagger[:, j]),
                    np.std(bp_buckle[:, j]),
                    np.std(bp_prop[:, j]),
                    np.std(bp_open[:, j]),
                    np.std(bp_shift[:, j]),
                    np.std(bp_slide[:, j]),
                    np.std(bp_rise[:, j]),
                    np.std(bp_tilt[:, j]),
                    np.std(bp_roll[:, j]),
                    np.std(bp_twist[:, j]),
                ]
            )
    else:
        # fmt: off
        bp_rise, bp_shift, bp_slide, bp_tilt, bp_roll, bp_twist = [], [], [], [], [], [], [], [], [], [], [], []
        # fmt: on
        for i in range(len(profile)):
            # print i
            bp_rise.append(profile.values()[i].Rise)
            bp_shift.append(profile.values()[i].Shift)
            bp_slide.append(profile.values()[i].Slide)
            bp_tilt.append(profile.values()[i].Tilt)
            bp_roll.append(profile.values()[i].Roll)
            bp_twist.append(profile.values()[i].Twist)
        bp_rise, bp_shift, bp_slide, bp_tilt, bp_roll, bp_twist = (
            np.array(bp_shear),
            np.array(bp_stretch),
            np.array(bp_stagger),
            np.array(bp_rise),
            np.array(bp_shift),
            np.array(bp_slide),
            np.array(bp_buckle),
            np.array(bp_prop),
            np.array(bp_open),
            np.array(bp_tilt),
            np.array(bp_roll),
            np.array(bp_twist),
        )
        na_avg, na_std = [], []
        for j in range(len(bp_shift[0])):
            na_avg.append(
                [
                    np.mean(bp_shift[:, j]),
                    np.mean(bp_slide[:, j]),
                    np.mean(bp_rise[:, j]),
                    np.mean(bp_tilt[:, j]),
                    np.mean(bp_roll[:, j]),
                    np.mean(bp_twist[:, j]),
                ]
            )
            na_std.append(
                [
                    np.std(bp_shift[:, j]),
                    np.std(bp_slide[:, j]),
                    np.std(bp_rise[:, j]),
                    np.std(bp_tilt[:, j]),
                    np.std(bp_roll[:, j]),
                    np.std(bp_twist[:, j]),
                ]
            )

    na_avg, na_std = np.array(na_avg), np.array(na_std)
    return na_avg, na_std


class BaseX3DNA(object):
    """Baseclass for X3DNA_ analysis, providing plotting and utility functions.

    When methods return helicoidal basepair parameter as lists, then the order
    is always

    ====== ==============
    index  parameter
    ====== ==============
     0     shear
     1     stretch
     2     stagger
     3     buckle
     4     propeller
     5     opening
     6     shift
     7     slide
     8     rise
     9     tilt
    10     roll
    11     twist
    ====== ==============

    for each nucleic acid pair.

    .. deprecated:: 2.7.0
        X3DNA will be removed in 3.0.0.

    .. _X3DNA: http://x3dna.org

    """

    def save(self, filename="x3dna.pickle"):
        """Save :attr:`profiles` as a Python pickle file *filename*.

        Load profiles dictionary with ::

           import cPickle
           profiles = cPickle.load(open(filename))

        """
        import cPickle

        cPickle.dump(
            self.profiles, open(filename, "wb"), cPickle.HIGHEST_PROTOCOL
        )

    def mean_std(self):
        """Returns the mean and standard deviation of base parameters.

        Returns
        -------
        (list, list)
            The tuple contains two lists with the means and the standard deviations
            for the helicoidal parameters. The order for both lists is ``[shear,
            stretch, stagger, buckle, propeller, opening, shift, slide, rise, tilt,
            roll, twist]``.
        """

        # fmt: off
        bp_shear, bp_stretch, bp_stagger, bp_rise, bp_shift, bp_slide, bp_buckle, bp_prop, bp_open, bp_tilt, bp_roll,\
            bp_twist = [], [], [], [], [], [], [], [], [], [], [], []
        # fmt: on
        for i in range(len(self.profiles)):
            bp_shear.append(self.profiles.values()[i].Shear)
            bp_stretch.append(self.profiles.values()[i].Stretch)
            bp_stagger.append(self.profiles.values()[i].Stagger)
            bp_buckle.append(self.profiles.values()[i].Buckle)
            bp_prop.append(self.profiles.values()[i].Propeller)
            bp_open.append(self.profiles.values()[i].Opening)
            bp_rise.append(self.profiles.values()[i].Rise)
            bp_shift.append(self.profiles.values()[i].Shift)
            bp_slide.append(self.profiles.values()[i].Slide)
            bp_tilt.append(self.profiles.values()[i].Tilt)
            bp_roll.append(self.profiles.values()[i].Roll)
            bp_twist.append(self.profiles.values()[i].Twist)
        # fmt: off
        bp_shear, bp_stretch, bp_stagger, bp_rise, bp_shift, bp_slide, bp_buckle, bp_prop, bp_open, bp_tilt, bp_roll,\
            bp_twist = np.array(bp_shear), np.array(bp_stretch), np.array(bp_stagger), np.array(bp_rise),\
            np.array(bp_shift), np.array(bp_slide), np.array(bp_buckle), np.array(bp_prop),\
            np.array(bp_open), np.array(bp_tilt), np.array(bp_roll), np.array(bp_twist)
        # fmt: on
        na_avg, na_std = [], []
        for j in range(len(bp_shear[0])):
            na_avg.append(
                [
                    np.mean(bp_shear[:, j]),
                    np.mean(bp_stretch[:, j]),
                    np.mean(bp_stagger[:, j]),
                    np.mean(bp_buckle[:, j]),
                    np.mean(bp_prop[:, j]),
                    np.mean(bp_open[:, j]),
                    np.mean(bp_shift[:, j]),
                    np.mean(bp_slide[:, j]),
                    np.mean(bp_rise[:, j]),
                    np.mean(bp_tilt[:, j]),
                    np.mean(bp_roll[:, j]),
                    np.mean(bp_twist[:, j]),
                ]
            )
            na_std.append(
                [
                    np.std(bp_shear[:, j]),
                    np.std(bp_stretch[:, j]),
                    np.std(bp_stagger[:, j]),
                    np.std(bp_buckle[:, j]),
                    np.std(bp_prop[:, j]),
                    np.std(bp_open[:, j]),
                    np.std(bp_shift[:, j]),
                    np.std(bp_slide[:, j]),
                    np.std(bp_rise[:, j]),
                    np.std(bp_tilt[:, j]),
                    np.std(bp_roll[:, j]),
                    np.std(bp_twist[:, j]),
                ]
            )
        na_avg, na_std = np.array(na_avg), np.array(na_std)
        return na_avg, na_std

    def mean(self):
        """Returns the mean value for the base parameters.

        Returns
        -------
        list
            The list contains the means for the helicoidal parameters. The
            order is ``[shear, stretch, stagger, buckle, propeller, opening,
            shift, slide, rise, tilt, roll, twist]``.

        """
        # fmt: off
        bp_shear, bp_stretch, bp_stagger, bp_rise, bp_shift, bp_slide, bp_buckle, bp_prop, bp_open, bp_tilt, bp_roll,\
            bp_twist = [], [], [], [], [], [], [], [], [], [], [], []
        # fmt: on
        for i in range(len(self.profiles)):
            bp_shear.append(self.profiles.values()[i].Shear)
            bp_stretch.append(self.profiles.values()[i].Stretch)
            bp_stagger.append(self.profiles.values()[i].Stagger)
            bp_buckle.append(self.profiles.values()[i].Buckle)
            bp_prop.append(self.profiles.values()[i].Propeller)
            bp_open.append(self.profiles.values()[i].Opening)
            bp_rise.append(self.profiles.values()[i].Rise)
            bp_shift.append(self.profiles.values()[i].Shift)
            bp_slide.append(self.profiles.values()[i].Slide)
            bp_tilt.append(self.profiles.values()[i].Tilt)
            bp_roll.append(self.profiles.values()[i].Roll)
            bp_twist.append(self.profiles.values()[i].Twist)
        # fmt: off
        bp_shear, bp_stretch, bp_stagger, bp_rise, bp_shift, bp_slide, bp_buckle, bp_prop, bp_open, bp_tilt, bp_roll,\
            bp_twist = np.array(bp_shear), np.array(bp_stretch), np.array(bp_stagger), np.array(bp_rise),\
            np.array(bp_shift), np.array(bp_slide), np.array(bp_buckle), np.array(bp_prop),\
            np.array(bp_open), np.array(bp_tilt), np.array(bp_roll), np.array(bp_twist)
        # fmt: on

        na_avg = []
        for j in range(len(bp_shear[0])):
            na_avg.append(
                [
                    np.mean(bp_shear[:, j]),
                    np.mean(bp_stretch[:, j]),
                    np.mean(bp_stagger[:, j]),
                    np.mean(bp_buckle[:j]),
                    np.mean(bp_prop[:, j]),
                    np.mean(bp_open[:, j]),
                    np.mean(bp_shift[:, j]),
                    np.mean(bp_slide[:, j]),
                    np.mean(bp_rise[:, j]),
                    np.mean(bp_tilt[:, j]),
                    np.mean(bp_roll[:, j]),
                    np.mean(bp_twist[:, j]),
                ]
            )
        na_avg = np.array(na_avg)
        return na_avg

    def std(self):
        """Returns the standard deviation for the base parameters.

        Returns
        -------
        list
            The list contains the standard deviations for the helicoidal
            parameters. The order is ``[shear, stretch, stagger, buckle,
            propeller, opening, shift, slide, rise, tilt, roll, twist]``.

        """
        # fmt: off
        bp_shear, bp_stretch, bp_stagger, bp_rise, bp_shift, bp_slide, bp_buckle, bp_prop, bp_open, bp_tilt, bp_roll,\
            bp_twist = [], [], [], [], [], [], [], [], [], [], [], []
        # fmt: on
        for i in range(len(self.profiles)):
            bp_shear.append(self.profiles.values()[i].Shear)
            bp_stretch.append(self.profiles.values()[i].Stretch)
            bp_stagger.append(self.profiles.values()[i].Stagger)
            bp_buckle.append(self.profiles.values()[i].Buckle)
            bp_prop.append(self.profiles.values()[i].Propeller)
            bp_open.append(self.profiles.values()[i].Opening)
            bp_rise.append(self.profiles.values()[i].Rise)
            bp_shift.append(self.profiles.values()[i].Shift)
            bp_slide.append(self.profiles.values()[i].Slide)
            bp_tilt.append(self.profiles.values()[i].Tilt)
            bp_roll.append(self.profiles.values()[i].Roll)
            bp_twist.append(self.profiles.values()[i].Twist)
        # fmt: off
        bp_shear, bp_stretch, bp_stagger, bp_rise, bp_shift, bp_slide, bp_buckle, bp_prop, bp_open, bp_tilt, bp_roll,\
            bp_twist = np.array(bp_shear), np.array(bp_stretch), np.array(bp_stagger), np.array(bp_rise),\
            np.array(bp_shift), np.array(bp_slide), np.array(bp_buckle), np.array(bp_prop),\
            np.array(bp_open), np.array(bp_tilt), np.array(bp_roll), np.array(bp_twist)
        # fmt: on

        na_std = []
        for j in range(len(bp_shear[0])):
            na_std.append(
                [
                    np.std(bp_shear[:, j]),
                    np.std(bp_stretch[:, j]),
                    np.std(bp_stagger[:, j]),
                    np.std(bp_buckle[:j]),
                    np.std(bp_prop[:, j]),
                    np.std(bp_open[:, j]),
                    np.std(bp_shift[:, j]),
                    np.std(bp_slide[:, j]),
                    np.std(bp_rise[:, j]),
                    np.std(bp_tilt[:, j]),
                    np.std(bp_roll[:, j]),
                    np.std(bp_twist[:, j]),
                ]
            )
        na_std = np.array(na_std)
        return na_std

    def plot(self, **kwargs):
        """Plot time-averaged base parameters for each basse pair in a 1D graph.

        One plot is produced for each parameter. It shows the the mean and
        standard deviation for each individual base pair. Each plot is saved to
        PNG file with name "<parameter_name>.png".

        Parameters
        ----------
        ax : matplotlib.pyplot.Axes (optional)
             Provide `ax` to have all plots plotted in the same axes.

        """

        na_avg, na_std = self.mean_std()
        for k in range(len(na_avg[0])):
            ax = kwargs.pop("ax", plt.subplot(111))
            x = list(range(1, len(na_avg[:, k]) + 1))
            ax.errorbar(x, na_avg[:, k], yerr=na_std[:, k], fmt="-o")
            ax.set_xlim(0, len(na_avg[:, k]) + 1)
            ax.set_xlabel(r"Nucleic Acid Number")
            param = self.profiles.values()[0].dtype.names[k]
            if param in [
                "Shear",
                "Stretch",
                "Stagger",
                "Rise",
                "Shift",
                "Slide",
            ]:
                ax.set_ylabel(r"{!s} ($\AA$)".format(param))
            else:
                ax.set_ylabel("{0!s} (deg)".format((param)))
            ax.figure.savefig("{0!s}.png".format((param)))
            ax.figure.clf()

    def sorted_profiles_iter(self):
        """Return an iterator over profiles sorted by frame/order parameter.

        The iterator produces tuples ``(q, profile)``. Typically, `q` is the
        frame number.
        """
        if self.profiles is None:
            return
        for q in sorted(self.profiles):
            yield (q, self.profiles[q])

    __iter__ = sorted_profiles_iter


class X3DNA(BaseX3DNA):
    """Run X3DNA_ on a single frame or a DCD trajectory.

    Only a subset of all X3DNA control parameters is supported and can be set
    with keyword arguments. For further details on X3DNA_ see the `X3DNA docs`_.

    Running X3DNA with the :class:`X3DNA` class is a 3-step process:

     1. set up the class with all desired parameters
     2. run X3DNA with :meth:`X3DNA.run`
     3. collect the data from the output file with :meth:`X3DNA.collect`

    The class also provides some simple plotting functions of the collected
    data such as :meth:`X3DNA.plot` or :meth:`X3DNA.plot3D`.

    When methods return helicoidal basepair parameter as lists, then the order
    is always

    ====== ==============
    index  parameter
    ====== ==============
     0     shear
     1     stretch
     2     stagger
     3     buckle
     4     propeller
     5     opening
     6     shift
     7     slide
     8     rise
     9     tilt
    10     roll
    11     twist
    ====== ==============

    .. versionadded:: 0.8

    .. deprecated:: 2.7.0
        X3DNA will be removed in 3.0.0.

    .. _`X3DNA docs`: http://forum.x3dna.org/
    """

    @deprecate(
        release="2.7.0",
        remove="3.0.0",
        message=(
            "X3DNA module is deprecated and will be removed in"
            "MDAnalysis 3.0.0, see #3788"
        ),
    )
    def __init__(self, filename, **kwargs):
        """Set up parameters to run X3DNA_ on PDB *filename*.

        Parameters
        ----------
        filename : str
               The `filename` is used as input for X3DNA in the
               :program:`xdna_ensemble` command.  It specifies the name of a
               PDB coordinate file to be used. This must be in Brookhaven
               protein databank format or something closely approximating
               this.
        executable : str (optional)
               Path to the :program:`xdna_ensemble` executable directories
               (e.g. ``/opt/x3dna/2.1 and /opt/x3dna/2.1/bin``) must be set
               and then added to export in bashrc file. See X3DNA
               documentation for set-up instructions.
        x3dna_param : bool (optional)
               Determines whether base step or base pair parameters will be
               calculated. If ``True`` (default) then stacked *base step*
               parameters will be analyzed.  If ``False`` then stacked *base
               pair* parameters will be analyzed.
        logfile : str (optional)
               Write output from X3DNA to `logfile` (default: "bp_step.par")


        See Also
        --------
        :class:`X3DNAtraj`
        """
        # list of temporary files, to be cleaned up on __del__
        self.tempfiles = [
            "auxiliary.par",
            "bestpairs.pdb",
            "bp_order.dat",
            "bp_helical.par",
            "cf_7methods.par",
            "col_chains.scr",
            "col_helices.scr",
            "hel_regions.pdb",
            "ref_frames.dat",
            "hstacking.pdb",
            "stacking.pdb",
        ]
        self.tempdirs = []
        self.filename = filename

        logger.info("Setting up X3DNA analysis for %(filename)r", vars(self))

        # guess executables
        self.exe = {}
        x3dna_exe_name = kwargs.pop("executable", "xdna_ensemble")
        self.x3dna_param = kwargs.pop("x3dna_param", True)
        self.exe["xdna_ensemble"] = which(x3dna_exe_name)
        if self.exe["xdna_ensemble"] is None:
            errmsg = "X3DNA binary {x3dna_exe_name!r} not found.".format(
                **vars()
            )
            logger.fatal(errmsg)
            logger.fatal(
                "%(x3dna_exe_name)r must be on the PATH or provided as keyword argument 'executable'.",
                vars(),
            )
            raise OSError(errno.ENOENT, errmsg)
        x3dnapath = os.path.dirname(self.exe["xdna_ensemble"])
        self.logfile = kwargs.pop("logfile", "bp_step.par")

        if self.x3dna_param is False:
            self.template = textwrap.dedent(
                """x3dna_ensemble analyze -b 355d.bps --one %(filename)r """
            )
        else:
            self.template = textwrap.dedent(
                """find_pair -s %(filename)r stdout |analyze stdin """
            )

        # sanity checks
        for program, path in self.exe.items():
            if path is None or which(path) is None:
                logger.error(
                    "Executable %(program)r not found, should have been %(path)r.",
                    vars(),
                )
        # results
        self.profiles = OrderedDict()

    def run(self, **kwargs):
        """Run X3DNA on the input file."""
        inpname = kwargs.pop("inpfile", None)
        outname = kwargs.pop("outfile", self.logfile)

        x3dnaargs = vars(self).copy()
        x3dnaargs.update(kwargs)
        x3dna_param = kwargs.pop("x3dna_param", self.x3dna_param)

        inp = self.template % x3dnaargs
        if inpname:
            with open(inpname, "w") as f:
                f.write(inp)
            logger.debug("Wrote X3DNA input file %r for inspection", inpname)

        logger.info(
            "Starting X3DNA on %(filename)r (trajectory: %(dcd)r)", x3dnaargs
        )
        logger.debug("%s", self.exe["xdna_ensemble"])
        with open(outname, "w") as output:
            x3dna = subprocess.call([inp], shell=True)
        with open(outname, "r") as output:
            # X3DNA is not very good at setting returncodes so check ourselves
            for line in output:
                if line.strip().startswith(("*** ERROR ***", "ERROR")):
                    x3dna.returncode = 255
                    break
        if x3dna.bit_length != 0:
            logger.fatal(
                "X3DNA Failure (%d). Check output %r",
                x3dna.bit_length,
                outname,
            )
        logger.info("X3DNA finished: output file %(outname)r", vars())

    def collect(self, **kwargs):
        """Parse the output from a X3DNA run into numpy recarrays.

        Can deal with outputs containing multiple frames.

        The method saves the result as :attr:`X3DNA.profiles`, a dictionary
        indexed by the frame number. Each entry is a
        :class:`np.recarray`.

        If the keyword `outdir` is supplied (e.g. ".") then each profile is
        saved to a gzipped data file.

        Parameters
        ----------
        run : str, int (optional
              identifier, free form [1]
        outdir : str (optional)
              save output data under `outdir`/`run` if set to any other
              value but ``None`` [``None``]

        """
        #        Shear    Stretch   Stagger   Buckle   Prop-Tw   Opening     Shift     Slide     Rise      Tilt        Roll      Twist
        # 0123456789.0123456789.0123456789.0123456789.0123456789.0123456789.123456789.123456789.123456789.123456789.123456789.123456789.123456789.
        #            11          22          33          44
        # T-A     -0.033    -0.176     0.158   -12.177    -8.979     1.440     0.000     0.000     0.000     0.000     0.000     0.000
        # C-G     -0.529     0.122    -0.002    -7.983   -10.083    -0.091    -0.911     1.375     3.213    -0.766    -4.065    41.492
        # only parse bp_step.par
        x3dna_output = kwargs.pop("x3dnaout", self.logfile)

        run = kwargs.pop("run", 1)  # id number
        outdir = kwargs.pop("outdir", os.path.curdir)

        logger.info("Collecting X3DNA profiles for run with id %s", run)
        length = 1  # length of trajectory --- is this really needed?? No... just for info
        if "*" in self.filename:
            import glob

            filenames = glob.glob(self.filename)
            length = len(filenames)
            if length == 0:
                logger.error(
                    "Glob pattern %r did not find any files.", self.filename
                )
                raise ValueError(
                    "Glob pattern {0!r} did not find any files.".format(
                        self.filename
                    )
                )
            logger.info(
                "Found %d input files based on glob pattern %s",
                length,
                self.filename,
            )

        # one recarray for each frame, indexed by frame number
        self.profiles = OrderedDict()

        logger.info(
            "Run %s: Reading %d X3DNA profiles from %r",
            run,
            length,
            x3dna_output,
        )
        x3dna_profile_no = 0
        records = []
        with open(x3dna_output, "r") as x3dna:
            read_data = False
            for line in x3dna:
                line = line.rstrip()  # preserve columns (FORTRAN output...)
                if self.x3dna_param is False:
                    if line.startswith("#        Shear"):
                        read_data = True
                        logger.debug("Started reading data")
                        fields = line.split()
                        x3dna_profile_no = int(
                            1
                        )  # useless int value code based off hole plugin
                        records = []
                        continue
                    if read_data:
                        if len(line.strip()) != 0:
                            try:
                                (
                                    Sequence,
                                    Shear,
                                    Stretch,
                                    Stagger,
                                    Buckle,
                                    Propeller,
                                    Opening,
                                    Shift,
                                    Slide,
                                    Rise,
                                    Tilt,
                                    Roll,
                                    Twist,
                                ) = line.split()
                            except:
                                logger.critical(
                                    "Run %d: Problem parsing line %r",
                                    run,
                                    line.strip(),
                                )
                                logger.exception(
                                    "Check input file %r.", x3dna_output
                                )
                                raise
                            records.append(
                                [
                                    float(Shear),
                                    float(Stretch),
                                    float(Stagger),
                                    float(Buckle),
                                    float(Propeller),
                                    float(Opening),
                                    float(Shift),
                                    float(Slide),
                                    float(Rise),
                                    float(Tilt),
                                    float(Roll),
                                    float(Twist),
                                ]
                            )
                            continue
                        else:
                            # end of records (empty line)
                            read_data = False
                else:
                    if line.startswith("#      Shift"):
                        read_data = True
                        logger.debug("Started reading data")
                        fields = line.split()
                        x3dna_profile_no = int(
                            1
                        )  # useless int value code based off hole plugin
                        records = []
                        continue
                    if read_data:
                        if len(line.strip()) != 0:
                            try:
                                (
                                    Sequence,
                                    Shift,
                                    Slide,
                                    Rise,
                                    Tilt,
                                    Roll,
                                    Twist,
                                ) = line.split()
                            except:
                                logger.critical(
                                    "Run %d: Problem parsing line %r",
                                    run,
                                    line.strip(),
                                )
                                logger.exception(
                                    "Check input file %r.", x3dna_output
                                )
                                raise
                            records.append(
                                [
                                    float(Shift),
                                    float(Slide),
                                    float(Rise),
                                    float(Tilt),
                                    float(Roll),
                                    float(Twist),
                                ]
                            )
                            continue
                        else:
                            # end of records (empty line)
                            read_data = False
            if self.x3dna_param is False:
                frame_x3dna_output = np.rec.fromrecords(
                    records,
                    formats="f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8",
                    names="Shear,Stretch,Stagger,Buckle,Propeller,Opening,"
                    "Shift,Slide,Rise,Tilt,Roll,Twist",
                )
            else:
                frame_x3dna_output = np.rec.fromrecords(
                    records,
                    formats="f8,f8,f8,f8,f8,f8",
                    names="Shift,Slide,Rise,Tilt,Roll,Twist",
                )
            # store the profile
            self.profiles[x3dna_profile_no] = frame_x3dna_output
            logger.debug(
                "Collected X3DNA profile for frame %d (%d datapoints)",
                x3dna_profile_no,
                len(frame_x3dna_output),
            )
            # save a profile for each frame (for debugging and scripted processing)
            # a tmp folder for each trajectory
            if outdir is not None:
                rundir = os.path.join(outdir, "run_" + str(run))
                os.system("rm -f tmp*.out")
                if not os.path.exists(rundir):
                    os.makedirs(rundir)
                frame_x3dna_txt = os.path.join(
                    rundir,
                    "bp_step_{0!s}_{1:04d}.dat.gz".format(
                        run, x3dna_profile_no
                    ),
                )
                np.savetxt(frame_x3dna_txt, frame_x3dna_output)
                logger.debug(
                    "Finished with frame %d, saved as %r",
                    x3dna_profile_no,
                    frame_x3dna_txt,
                )
                # if we get here then we haven't found anything interesting
        if len(self.profiles) == length:
            logger.info(
                "Collected X3DNA profiles for %d frames", len(self.profiles)
            )
        else:
            logger.warning(
                "Missing data: Found %d X3DNA profiles from %d frames.",
                len(self.profiles),
                length,
            )

    def __del__(self):
        for f in self.tempfiles:
            try:
                os.unlink(f)
            except OSError:
                pass
        for d in self.tempdirs:
            shutil.rmtree(d, ignore_errors=True)


class X3DNAtraj(BaseX3DNA):
    """Analyze all frames in a trajectory.

    The :class:`X3DNA` class provides a direct interface to X3DNA. X3DNA itself
    has limited support for analysing trajectories but cannot deal with all the
    trajectory formats understood by MDAnalysis. This class can take any
    universe and feed it to X3DNA. By default it sequentially creates a PDB for
    each frame and runs X3DNA on the frame.


    .. deprecated:: 2.7.0
        X3DNA will be removed in 3.0.0.
    """

    @deprecate(
        release="2.7.0",
        remove="3.0.0",
        message=(
            "X3DNA module is deprecated and will be removed in"
            "MDAnalysis 3.0.0, see #3788"
        ),
    )
    def __init__(self, universe, **kwargs):
        """Set up the class.

        Parameters
        ----------
        universe : Universe
               The input trajectory as part of a
               :class:`~MDAnalysis.core.universe.Universe`; the trajectory is
               converted to a sequence of PDB files and X3DNA is run on each
               individual file. (Use the `start`, `stop`, and `step` keywords
               to slice the trajectory.)
        selection : str (optional)
               MDAnalysis selection string (default: "nucleic") to select the
               atoms that should be analyzed.
        start : int (optional)
        stop : int (optional)
        step : int (optional)
               frame indices to slice the trajectory as
               ``universe.trajectory[start, stop, step]``; by default, the whole
               trajectory is analyzed.
        x3dna_param : bool (optional)
               indicates whether stacked bases or stacked base-pairs will be
               analyzed. ``True`` is bases and ``False`` is stacked base-pairs
               [Default is ``True``].
        kwargs : keyword arguments (optional)
               All other keywords are passed on to :class:`X3DNA` (see there
               for description).


        See Also
        --------
        :class:`X3DNA`

        """
        self.universe = universe
        self.selection = kwargs.pop("selection", "nucleic")

        self.x3dna_param = kwargs.pop("x3dna_param", True)
        self.start = kwargs.pop("start", None)
        self.stop = kwargs.pop("stop", None)
        self.step = kwargs.pop("step", None)

        self.x3dna_kwargs = kwargs

        # processing

    def run(self, **kwargs):
        """Run X3DNA on the whole trajectory and collect profiles.

        Keyword arguments `start`, `stop`, and `step` can be used to only
        analyse part of the trajectory. The defaults are the values provided to
        the class constructor.
        """
        start = kwargs.pop("start", self.start)
        stop = kwargs.pop("stop", self.stop)
        step = kwargs.pop("step", self.step)
        x3dna_param = kwargs.pop("x3dna_param", self.x3dna_param)
        x3dna_kw = self.x3dna_kwargs.copy()
        x3dna_kw.update(kwargs)

        profiles = OrderedDict()

        nucleic = self.universe.select_atoms(self.selection)
        for ts in self.universe.trajectory[start:stop:step]:
            logger.info("X3DNA analysis frame %4d ", ts.frame)
            fd, pdbfile = tempfile.mkstemp(suffix=".pdb")
            os.close(fd)
            nucleic.write(pdbfile)
            os.system("find_pair {0!s} 355d.bps".format(pdbfile))
            try:
                nucleic.write(pdbfile)
                x3dna_profiles = self.run_x3dna(pdbfile, **x3dna_kw)
            finally:
                try:
                    os.unlink(pdbfile)
                except OSError:
                    pass
            if len(x3dna_profiles) != 1:
                err_msg = "Got {0} profiles ({1}) --- should be 1 (time step {2})".format(
                    len(x3dna_profiles), x3dna_profiles.keys(), ts
                )
                logger.error(err_msg)
                warnings.warn(err_msg)
            profiles[ts.frame] = x3dna_profiles.values()[0]
        self.profiles = profiles

    def run_x3dna(self, pdbfile, **kwargs):
        """Run X3DNA on a single PDB file `pdbfile`."""
        kwargs["x3dna_param"] = self.x3dna_param
        H = X3DNA(pdbfile, **kwargs)
        H.run()
        H.collect()
        return H.profiles



import numpy as np
from numpy.testing import assert_equal

from MDAnalysis.topology.base import squash_by, change_squash


class TestSquash(object):
    atom_resids = np.array([2, 2, 1, 1, 5, 5, 4, 4])
    atom_resnames = np.array(
        ["A", "A", "B", "B", "C", "C", "D", "D"], dtype=object
    )

    def test_squash(self):
        atom_residx, resids, (resnames,) = squash_by(
            self.atom_resids, self.atom_resnames
        )

        assert_equal(atom_residx, np.array([1, 1, 0, 0, 3, 3, 2, 2]))
        assert_equal(resids, np.array([1, 2, 4, 5]))
        assert_equal(resnames, np.array(["B", "A", "D", "C"]))


class TestChangeSquash(object):
    def test_resid_squash(self):
        # System of 6 atoms in 3 residues
        # Residues 1 & 2 are Segid A, Residue 3 is Segid B
        # Resid 2 is repeated twice! Should be detected as 2 distinct residues
        resids = np.array([2, 2, 3, 3, 2, 2])
        resnames = np.array(["RsA", "RsA", "RsB", "RsB", "RsC", "RsC"])
        segids = np.array(["A", "A", "A", "A", "B", "B"])

        residx, (new_resids, new_resnames, new_segids) = change_squash(
            (resids,), (resids, resnames, segids)
        )

        assert_equal(residx, np.array([0, 0, 1, 1, 2, 2]))
        assert_equal(new_resids, np.array([2, 3, 2]))
        assert_equal(new_resnames, np.array(["RsA", "RsB", "RsC"]))
        assert_equal(new_segids, np.array(["A", "A", "B"]))

    def test_segid_squash(self):
        segids = np.array(["A", "A", "B"])

        segidx, (new_segids,) = change_squash((segids,), (segids,))

        assert_equal(segidx, np.array([0, 0, 1]))
        assert_equal(new_segids, np.array(["A", "B"]))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
dimensionality reduction frontend --- :mod:`MDAnalysis.analysis.encore.clustering.DimensionalityReductionMethod`
================================================================================================================

The module defines classes for interfacing to various dimensionality reduction
algorithms. One has been implemented natively, and will always be available,
while others are available only if scikit-learn is installed

:Author: Matteo Tiberti, Wouter Boomsma, Tone Bengtsen

.. versionadded:: 0.16.0

.. deprecated:: 2.8.0
   This module is deprecated in favour of the 
   MDAKit `mdaencore <https://mdanalysis.org/mdaencore/>`_ and will be removed
   in MDAnalysis 3.0.0.

"""
import logging
import warnings

# Import native affinity propagation implementation
from . import stochasticproxembed

# Attempt to import scikit-learn clustering algorithms
try:
    import sklearn.decomposition
except ImportError:
    sklearn = None
    import warnings

    warnings.warn(
        "sklearn.decomposition could not be imported: some "
        "functionality will not be available in "
        "encore.dimensionality_reduction()",
        category=ImportWarning,
    )


class DimensionalityReductionMethod(object):
    """
    Base class for any Dimensionality Reduction Method
    """

    # Whether the method accepts a distance matrix
    accepts_distance_matrix = True

    def __call__(self, x):
        """
        Parameters
        ----------

        x
            either trajectory coordinate data (np.array) or an
            encore.utils.TriangularMatrix, encoding the conformational
            distance matrix


        Returns
        -------
        numpy.array
            coordinates in reduced space

        """
        raise NotImplementedError(
            "Class {0} doesn't implement __call__()".format(
                self.__class__.__name__
            )
        )


class StochasticProximityEmbeddingNative(DimensionalityReductionMethod):
    """
    Interface to the natively implemented Affinity propagation procedure.
    """

    def __init__(
        self,
        dimension=2,
        distance_cutoff=1.5,
        min_lam=0.1,
        max_lam=2.0,
        ncycle=100,
        nstep=10000,
    ):
        """
        Parameters
        ----------

        dimension : int
            Number of dimensions to which the conformational space will be
            reduced to (default is 3).

        min_lam : float, optional
            Final lambda learning rate (default is 0.1).

        max_lam : float, optional
            Starting lambda learning rate parameter (default is 2.0).

        ncycle : int, optional
            Number of cycles per run (default is 100). At the end of every
            cycle, lambda is updated.

        nstep : int, optional
            Number of steps per cycle (default is 10000)

        """
        self.dimension = dimension
        self.distance_cutoff = distance_cutoff
        self.min_lam = min_lam
        self.max_lam = max_lam
        self.ncycle = ncycle
        self.nstep = nstep
        self.stressfreq = -1

    def __call__(self, distance_matrix):
        """
        Parameters
        ----------

        distance_matrix : encore.utils.TriangularMatrix
            conformational distance matrix


        Returns
        -------
        numpy.array
            coordinates in reduced space

        """
        final_stress, coordinates = (
            stochasticproxembed.StochasticProximityEmbedding(
                s=distance_matrix,
                rco=self.distance_cutoff,
                dim=self.dimension,
                minlam=self.min_lam,
                maxlam=self.max_lam,
                ncycle=self.ncycle,
                nstep=self.nstep,
                stressfreq=self.stressfreq,
            )
        )
        return coordinates, {"final_stress": final_stress}


if sklearn:

    class PrincipalComponentAnalysis(DimensionalityReductionMethod):
        """
        Interface to the PCA dimensionality reduction method implemented in
        sklearn.
        """

        # Whether the method accepts a distance matrix
        accepts_distance_matrix = False

        def __init__(self, dimension=2, **kwargs):
            """
            Parameters
            ----------

            dimension : int
                Number of dimensions to which the conformational space will be
                reduced to (default is 3).
            """
            self.pca = sklearn.decomposition.PCA(
                n_components=dimension, **kwargs
            )

        def __call__(self, coordinates):
            """
            Parameters
            ----------

            coordinates : np.array
                trajectory atom coordinates


            Returns
            -------
            numpy.array
                coordinates in reduced space
            """
            coordinates = self.pca.fit_transform(coordinates)
            return coordinates.T, {}



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from io import StringIO

import numpy as np
from numpy.testing import assert_equal, assert_allclose
import pytest

import MDAnalysis as mda
from MDAnalysisTests.topology.base import ParserBase

from MDAnalysisTests.datafiles import (
    mol2_molecule,
    mol2_molecules,
)


mol2_wo_opt_col = """\
@<TRIPOS>MOLECULE
mol2_wo_opt_col
2
SMALL
NO_CHARGES
@<TRIPOS>ATOM
  1 N1       6.8420     9.9900    22.7430 N.am
  2 N2       4.4000     9.1300    20.4710 N.am
"""

mol2_partial_opt_col = """\
@<TRIPOS>MOLECULE
mol2_partial_opt_col
2
SMALL
NO_CHARGES
@<TRIPOS>ATOM
  1 N1       6.8420     9.9900    22.7430 N.am  1
  2 N2       4.4000     9.1300    20.4710 N.am  2
"""

mol2_resname_unformat = """\
@<TRIPOS>MOLECULE
mol2_resname_unformat
2 1 0 0 0
SMALL
NO_CHARGES

@<TRIPOS>ATOM
  1 O        3.0000    0.0000    0.0000 O.3     1
  2 C        2.0000    0.0000    0.0000 C.2     1  UNL1
"""

mol2_wo_required_col = """\
@<TRIPOS>MOLECULE
mol2_wo_required_col
2
SMALL
NO_CHARGES
@<TRIPOS>ATOM
  1 N1       6.8420     9.9900    22.7430
  2 N2       4.4000     9.1300    20.4710
"""

mol2_no_charge_error1 = """\
@<TRIPOS>MOLECULE
mol2_no_charge_error1
2 1 0 0 0
SMALL
NO_CHARGES

@<TRIPOS>ATOM
  1 O        3.0000    0.0000    0.0000 O.3     1  UNL1       -0.0520
  2 C        2.0000    0.0000    0.0000 C.2     1  UNL1        0.0520
"""

mol2_no_charge_error2 = """\
@<TRIPOS>MOLECULE
mol2_no_charge_error2
2 1 0 0 0
SMALL
MMFF94_CHARGES

@<TRIPOS>ATOM
  1 O        3.0000     0.0000    0.0000 O.3    1  UNL1
  2 C        2.0000     0.0000    0.0000 C.2    1  UNL1
"""

mol2_wrong_element = """\
@<TRIPOS>MOLECULE
mol2_wrong_element
3 0 0 0 0
SMALL
USER_CHARGES


@<TRIPOS>ATOM
  1 N1       6.8420     9.9900    22.7430 N.am  1 Q101  -0.8960
  2 S1       8.1400     9.2310    23.3330 X.o2  1 Q101   1.3220
  3 N2       4.4000     9.1300    20.4710 XX.am  1 Q101  -0.3970
"""

mol2_all_wrong_elements = """\
@<TRIPOS>MOLECULE
mol2_all_wrong_elements
3 0 0 0 0
SMALL
USER_CHARGES


@<TRIPOS>ATOM
  1 N1       6.8420     9.9900    22.7430 X.am  1 Q101  -0.8960
  2 S1       8.1400     9.2310    23.3330 X.o2  1 Q101   1.3220
  3 N2       4.4000     9.1300    20.4710 XX.am  1 Q101  -0.3970
"""


mol2_fake = """\
@<TRIPOS>MOLECULE
mol2_fake
26 0 0 0 0
SMALL
USER_CHARGES


@<TRIPOS>ATOM
  1 H1       0.0000     1.0000    10.0000 H.spc  1 XXXX   5.0000
  2 H2       0.0000     1.0000    10.0000 H.t3p  1 XXXX   5.0000
  3 H3       0.0000     1.0000    10.0000 H.xyz  1 XXXX   5.0000
  4 C1       0.0000     1.0000    10.0000 C.1  1 XXXX   5.0000
  5 C2       0.0000     1.0000    10.0000 C.2  1 XXXX   5.0000
  5 C3       0.0000     1.0000    10.0000 C.3  1 XXXX   5.0000
  6 C4       0.0000     1.0000    10.0000 C.ar  1 XXXX   5.0000
  7 C5       0.0000     1.0000    10.0000 C.cat  1 XXXX   5.0000
  8 C6       0.0000     1.0000    10.0000 C.xyz  1 XXXX   5.0000
  9 N1       0.0000     1.0000    10.0000 N.1  1 XXXX   5.0000
  10 N2       0.0000     1.0000    10.0000 N.2  1 XXXX   5.0000
  11 N3       0.0000     1.0000    10.0000 N.3  1 XXXX   5.0000
  12 N4       0.0000     1.0000    10.0000 N.ar  1 XXXX   5.0000
  13 O1       0.0000     1.0000    10.0000 O.2  1 XXXX   5.0000
  14 O2       0.0000     1.0000    10.0000 O.3  1 XXXX   5.0000
  15 O3       0.0000     1.0000    10.0000 O.co2  1 XXXX   5.0000
  16 O4       0.0000     1.0000    10.0000 O.spc  1 XXXX   5.0000
  16 O5       0.0000     1.0000    10.0000 O.t3p  1 XXXX   5.0000
  17 S1       0.0000     1.0000    10.0000 S.3  1 XXXX   5.0000
  18 S2       0.0000     1.0000    10.0000 S.2  1 XXXX   5.0000
  19 S3       0.0000     1.0000    10.0000 S.O  1 XXXX   5.0000
  20 S4       0.0000     1.0000    10.0000 S.O2  1 XXXX   5.0000
  21 S5       0.0000     1.0000    10.0000 S.o  1 XXXX   5.0000
  22 S6       0.0000     1.0000    10.0000 S.o2  1 XXXX   5.0000
  23 P1       0.0000     1.0000    10.0000 P.3  1 XXXX   5.0000
  24 Cr1       0.0000     1.0000    10.0000 Cr.th  1 XXXX   5.0000
  25 Cr2       0.0000     1.0000    10.0000 Cr.oh  1 XXXX   5.0000
  26 Co1       0.0000     1.0000    10.0000 Co.oh  1 XXXX   5.0000
"""


class TestMOL2Base(ParserBase):
    parser = mda.topology.MOL2Parser.MOL2Parser
    expected_attrs = [
        "ids",
        "names",
        "types",
        "charges",
        "resids",
        "resnames",
        "bonds",
        "elements",
    ]

    guessed_attrs = ["masses"]
    expected_n_atoms = 49
    expected_n_residues = 1
    expected_n_segments = 1

    def test_attr_size(self, top):
        assert len(top.ids) == top.n_atoms
        assert len(top.names) == top.n_atoms
        assert len(top.types) == top.n_atoms
        assert len(top.charges) == top.n_atoms
        assert len(top.resids) == top.n_residues
        assert len(top.resnames) == top.n_residues
        assert len(top.elements) == top.n_atoms

    def test_bonds(self, top):
        assert len(top.bonds) == 49  # bonds for 49 atoms
        assert len(top.bonds.values) == 51  # this many bonds

    @pytest.fixture(params=[mol2_molecule, mol2_molecules])
    def filename(self, request):
        return request.param


def test_bond_orders():
    ref_orders = (
        "am 1 1 2 1 2 1 1 am 1 1 am 2 2 "
        "1 1 1 1 1 1 1 1 1 1 1 1 1 1 "
        "ar ar ar 1 ar 1 ar 1 ar 1 1 1 "
        "2 1 1 1 1 2 1 1 2 1 1"
    ).split()
    u = mda.Universe(mol2_molecule)
    orders = [bond.order for bond in u.atoms.bonds]
    assert_equal(orders, ref_orders)


def test_elements():
    u = mda.Universe(mol2_molecule)

    assert_equal(
        u.atoms.elements[:5], np.array(["N", "S", "N", "N", "O"], dtype="U3")
    )


# Test for #2927
def test_elements_selection():
    u = mda.Universe(mol2_molecule)
    ag = u.select_atoms("element S")

    assert_equal(ag.elements, np.array(["S", "S"], dtype="U3"))


def test_wrong_elements_warnings():
    with pytest.warns(UserWarning, match="Unknown elements found") as record:
        u = mda.Universe(StringIO(mol2_wrong_element), format="MOL2")

    # One warning from invalid elements, one from masses PendingDeprecationWarning
    assert len(record) == 3

    expected_elements = np.array(["N", "", ""], dtype=object)
    guseed_masses = np.array([14.007, 0.0, 0.0], dtype=float)
    gussed_types = np.array(["N.am", "X.o2", "XX.am"])

    assert_equal(u.atoms.elements, expected_elements)
    assert_equal(u.atoms.types, gussed_types)
    assert_allclose(u.atoms.masses, guseed_masses)


def test_all_wrong_elements_warnings():
    with pytest.warns(UserWarning, match="Unknown elements found"):
        u = mda.Universe(StringIO(mol2_all_wrong_elements), format="MOL2")

    with pytest.raises(
        mda.exceptions.NoDataError,
        match="This Universe does not contain element " "information",
    ):

        u.atoms.elements


def test_all_elements():
    with pytest.warns(UserWarning, match="Unknown elements found"):
        u = mda.Universe(StringIO(mol2_fake), format="MOL2")

    expected = (
        ["H"] * 2
        + [""]
        + ["C"] * 5
        + [""]
        + ["N"] * 4
        + ["O"] * 5
        + ["S"] * 6
        + ["P"]
        + ["Cr"] * 2
        + ["Co"]
    )
    expected = np.array(expected, dtype=object)
    assert_equal(u.atoms.elements, expected)


# Test for Issue #3385 / PR #3598
def test_wo_optional_columns():
    u = mda.Universe(StringIO(mol2_wo_opt_col), format="MOL2")
    assert_equal(u.atoms.resids, np.array([1, 1]))
    with pytest.raises(mda.exceptions.NoDataError):
        u.atoms.resnames
    with pytest.raises(mda.exceptions.NoDataError):
        u.atoms.charges


def test_partial_optional_columns():
    u = mda.Universe(StringIO(mol2_partial_opt_col), format="MOL2")
    assert_equal(u.atoms.resids, np.array([1, 2]))
    with pytest.raises(mda.exceptions.NoDataError):
        u.atoms.resnames
    with pytest.raises(mda.exceptions.NoDataError):
        u.atoms.charges


def test_mol2_wo_required_columns():
    with pytest.raises(
        ValueError, match="The @<TRIPOS>ATOM block in mol2 file"
    ):
        u = mda.Universe(StringIO(mol2_wo_required_col), format="MOL2")


def test_mol2_no_charges():
    with pytest.raises(ValueError, match="indicates no charges"):
        u = mda.Universe(StringIO(mol2_no_charge_error1), format="MOL2")
    with pytest.raises(ValueError, match="indicates a charge model"):
        u = mda.Universe(StringIO(mol2_no_charge_error2), format="MOL2")


def test_unformat():
    with pytest.raises(ValueError, match="Some atoms in the mol2 file"):
        u = mda.Universe(StringIO(mol2_resname_unformat), format="MOL2")


def test_guessed_masses():
    u = mda.Universe(mol2_molecules)
    assert_allclose(
        u.atoms.masses[:7],
        [14.007, 32.06, 14.007, 14.007, 15.999, 15.999, 12.011],
    )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
from numpy.testing import assert_equal, assert_allclose

import MDAnalysis as mda

from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import (
    GMS_ASYMOPT,  # c1opt.gms.gz
    GMS_SYMOPT,  # symopt.gms
    GMS_ASYMSURF,  # surf2wat.gms
)


class GMSBase(ParserBase):
    parser = mda.topology.GMSParser.GMSParser
    expected_attrs = ["names", "atomiccharges"]
    guessed_attrs = ["masses", "types"]
    expected_n_residues = 1
    expected_n_segments = 1


class TestGMSASYMOPT(GMSBase):
    expected_n_atoms = 6
    ref_filename = GMS_ASYMOPT

    def test_names(self, top):
        assert_equal(top.names.values, ["O", "H", "H", "O", "H", "H"])

    def test_types(self, top):
        assert_equal(top.atomiccharges.values, [8, 1, 1, 8, 1, 1])

    def test_guessed_masses(self, filename):
        u = mda.Universe(filename)
        expected = [15.999, 1.008, 1.008, 15.999, 1.008, 1.008]
        assert_allclose(u.atoms.masses, expected)

    def test_guessed_types(self, filename):
        u = mda.Universe(filename)
        expected = ["O", "H", "H", "O", "H", "H"]
        assert (u.atoms.types == expected).all()


class TestGMSSYMOPT(GMSBase):
    expected_n_atoms = 4
    ref_filename = GMS_SYMOPT

    def test_names(self, top):
        assert_equal(
            top.names.values, ["CARBON", "CARBON", "HYDROGEN", "HYDROGEN"]
        )

    def test_types(self, top):
        assert_equal(top.atomiccharges.values, [6, 6, 1, 1])


class TestGMSASYMSURF(TestGMSASYMOPT):
    ref_filename = GMS_ASYMSURF



import pytest
import glob
import MDAnalysis as mda

from MDAnalysis.analysis.dssp import DSSP, translate
from MDAnalysisTests.datafiles import DSSP as DSSP_FOLDER
from MDAnalysisTests.datafiles import TPR, XTC


# Files that match glob pattern '????.pdb.gz' and matching '????.pdb.dssp' files,
# containing the secondary structure assignment string, will be tested automatically.
@pytest.mark.parametrize(
    "pdb_filename", glob.glob(f"{DSSP_FOLDER}/?????.pdb.gz")
)
def test_file_guess_hydrogens(pdb_filename, client_DSSP):
    u = mda.Universe(pdb_filename)
    with open(f"{pdb_filename.rstrip('.gz')}.dssp", "r") as fin:
        correct_answ = fin.read().strip().split()[0]

    run = DSSP(u, guess_hydrogens=True).run(**client_DSSP)
    answ = "".join(run.results.dssp[0])
    assert answ == correct_answ


def test_trajectory(client_DSSP):
    u = mda.Universe(TPR, XTC).select_atoms("protein").universe
    run = DSSP(u).run(**client_DSSP, stop=10)
    first_frame = "".join(run.results.dssp[0])
    last_frame = "".join(run.results.dssp[-1])
    avg_frame = "".join(translate(run.results.dssp_ndarray.mean(axis=0)))

    assert (
        first_frame[:10] != last_frame[:10] == avg_frame[:10] == "-EEEEEE---"
    )
    protein = mda.Universe(TPR, XTC).select_atoms("protein")
    run = DSSP(protein).run(**client_DSSP, stop=10)


def test_atomgroup(client_DSSP):
    protein = mda.Universe(TPR, XTC).select_atoms("protein")
    run = DSSP(protein).run(**client_DSSP, stop=10)
    first_frame = "".join(run.results.dssp[0])
    last_frame = "".join(run.results.dssp[-1])
    avg_frame = "".join(translate(run.results.dssp_ndarray.mean(axis=0)))

    assert (
        first_frame[:10] != last_frame[:10] == avg_frame[:10] == "-EEEEEE---"
    )


def test_trajectory_with_hydrogens(client_DSSP):
    u = mda.Universe(TPR, XTC).select_atoms("protein").universe
    run = DSSP(u, guess_hydrogens=False).run(**client_DSSP, stop=10)
    first_frame = "".join(run.results.dssp[0])
    last_frame = "".join(run.results.dssp[-1])
    avg_frame = "".join(translate(run.results.dssp_ndarray.mean(axis=0)))

    assert (
        first_frame[:10] == last_frame[:10] == avg_frame[:10] == "-EEEEEE---"
    )


@pytest.mark.parametrize(
    "pdb_filename", glob.glob(f"{DSSP_FOLDER}/2xdgA.pdb.gz")
)
def test_trajectory_without_hydrogen_fails(pdb_filename, client_DSSP):
    u = mda.Universe(pdb_filename)
    with pytest.raises(ValueError):
        DSSP(u, guess_hydrogens=False).run(**client_DSSP)


@pytest.mark.parametrize(
    "pdb_filename", glob.glob(f"{DSSP_FOLDER}/1mr1D_failing.pdb.gz")
)
def test_trajectory_with_uneven_number_of_atoms_fails(
    pdb_filename, client_DSSP
):
    u = mda.Universe(pdb_filename)
    with pytest.raises(ValueError):
        DSSP(u, guess_hydrogens=True).run(**client_DSSP)


@pytest.mark.parametrize(
    "pdb_filename", glob.glob(f"{DSSP_FOLDER}/wrong_hydrogens.pdb.gz")
)
def test_exception_raises_with_atom_index(pdb_filename, client_DSSP):
    u = mda.Universe(pdb_filename)
    with pytest.raises(
        ValueError,
        match="Residue <Residue SER, 298> contains*",
    ):
        DSSP(u, guess_hydrogens=False).run(**client_DSSP)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
GRO file format --- :mod:`MDAnalysis.coordinates.GRO`
======================================================

Classes to read and write Gromacs_ GRO_ coordinate files; see the notes on the
`GRO format`_ which includes a conversion routine for the box.


Writing GRO files
-----------------

By default any written GRO files will renumber the atom ids to move sequentially
from 1.  This can be disabled, and instead the original atom ids kept, by
using the `reindex=False` keyword argument.  This is useful when writing a
subsection of a larger Universe while wanting to preserve the original
identities of atoms.

For example::

   >>> u = mda.Universe()`

   >>> u.atoms.write('out.gro', reindex=False)

   # OR
   >>> with mda.Writer('out.gro', reindex=False) as w:
   ...     w.write(u.atoms)


Classes
-------

.. autoclass:: GROReader
   :members:

.. autoclass:: GROWriter
   :members:


Developer notes: ``GROWriter`` format strings
---------------------------------------------

The :class:`GROWriter` class has a :attr:`GROWriter.fmt` attribute, which is a dictionary of different
strings for writing lines in ``.gro`` files.  These are as follows:

``n_atoms``
  For the first line of the gro file, supply the number of atoms in the system.
  E.g.::

      fmt['n_atoms'].format(42)

``xyz``
  An atom line without velocities.  Requires that the 'resid', 'resname',
  'name', 'index' and 'pos' keys be supplied.
  E.g.::

     fmt['xyz'].format(resid=1, resname='SOL', name='OW2', index=2, pos=(0.0, 1.0, 2.0))

``xyz_v``
  As above, but with velocities.  Needs an additional keyword 'vel'.

``box_orthorhombic``
  The final line of the gro file which gives box dimensions.  Requires
  the box keyword to be given, which should be the three cartesian dimensions.
  E.g.::

     fmt['box_orthorhombic'].format(box=(10.0, 10.0, 10.0))

``box_triclinic``
  As above, but for a non orthorhombic box. Requires the box keyword, but this
  time as a length 9 vector.  This is a flattened version of the (3,3) triclinic
  vector representation of the unit cell.  The rearrangement into the odd
  gromacs order is done automatically.


.. _Gromacs: http://www.gromacs.org
.. _GRO: https://manual.gromacs.org/current/reference-manual/file-formats.html#gro
.. _GRO format: http://chembytes.wikidot.com/g-grofile

"""
import re

import itertools
import warnings

import numpy as np

from . import base
from .core import triclinic_box, triclinic_vectors
from ..exceptions import NoDataError
from ..lib import util
from .timestep import Timestep

"""unitcell dimensions (A, B, C, alpha, beta, gamma)

GRO::

  8.00170   8.00170   5.65806   0.00000   0.00000   0.00000   0.00000   4.00085   4.00085

PDB::

  CRYST1   80.017   80.017   80.017  60.00  60.00  90.00 P 1           1

XTC: c.trajectory.ts._unitcell::

  array([[ 80.00515747,   0.        ,   0.        ],
         [  0.        ,  80.00515747,   0.        ],
         [ 40.00257874,  40.00257874,  56.57218552]], dtype=float32)
"""
# unit cell line (from http://manual.gromacs.org/current/online/gro.html)
# v1(x) v2(y) v3(z) v1(y) v1(z) v2(x) v2(z) v3(x) v3(y)
# 0     1     2      3     4     5     6    7     8
# This information now stored as _ts_order_x/y/z to keep DRY
_TS_ORDER_X = [0, 3, 4]
_TS_ORDER_Y = [5, 1, 6]
_TS_ORDER_Z = [7, 8, 2]


def _gmx_to_dimensions(box):
    # convert gromacs ordered box to [lx, ly, lz, alpha, beta, gamma] form
    x = box[_TS_ORDER_X]
    y = box[_TS_ORDER_Y]
    z = box[_TS_ORDER_Z]  # this ordering is correct! (checked it, OB)
    return triclinic_box(x, y, z)


class GROReader(base.SingleFrameReaderBase):
    """Reader for the Gromacs GRO structure format.

    .. note::
       This Reader will only read the first frame present in a file.


    .. note::
       GRO files with zeroed 3 entry unit cells (i.e. ``0.0   0.0   0.0``)
       are read as missing unit cell information. In these cases ``dimensions``
       will be set to ``None``.

    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based
    .. versionchanged:: 2.0.0
       Reader now only parses boxes defined with 3 or 9 fields.
       Reader now reads a 3 entry zero unit cell (i.e. ``[0, 0, 0]``) as a
       being without dimension information (i.e. will the timestep dimension
       to ``None``).
    """

    format = "GRO"
    units = {"time": None, "length": "nm", "velocity": "nm/ps"}
    _Timestep = Timestep

    def _read_first_frame(self):
        with util.openany(self.filename, "rt") as grofile:
            # Read first two lines to get number of atoms
            grofile.readline()
            self.n_atoms = n_atoms = int(grofile.readline())
            self.ts = ts = self._Timestep(n_atoms, **self._ts_kwargs)
            # Always try, and maybe add them later
            velocities = np.zeros((n_atoms, 3), dtype=np.float32)
            missed_vel = False

            # and the third line to get the spacing between coords (cs)
            # (dependent upon the GRO file precision)
            first_atomline = grofile.readline()
            cs = first_atomline[25:].find(".") + 1
            ts._pos[0] = [
                first_atomline[20 + cs * i : 20 + cs * (i + 1)]
                for i in range(3)
            ]
            try:
                velocities[0] = [
                    first_atomline[20 + cs * i : 20 + cs * (i + 1)]
                    for i in range(3, 6)
                ]
            except ValueError:
                # Remember that we got this error
                missed_vel = True

            # TODO: Handle missing unitcell?
            for pos, line in enumerate(grofile, start=1):
                # 2 header lines, 1 box line at end
                if pos == n_atoms:
                    try:
                        unitcell = np.float32(line.split())
                    except ValueError:
                        # Try to parse floats with 5 digits if no spaces between values...
                        unitcell = np.float32(
                            re.findall(r"(\d+\.\d{5})", line)
                        )
                    break

                ts._pos[pos] = [
                    line[20 + cs * i : 20 + cs * (i + 1)] for i in range(3)
                ]
                try:
                    velocities[pos] = [
                        line[20 + cs * i : 20 + cs * (i + 1)]
                        for i in range(3, 6)
                    ]
                except ValueError:
                    # Remember that we got this error
                    missed_vel = True

        if np.any(velocities):
            ts.velocities = velocities
            if missed_vel:
                warnings.warn(
                    "Not all velocities were present.  "
                    "Unset velocities set to zero."
                )

        self.ts.frame = 0  # 0-based frame number

        if len(unitcell) == 3:
            # special case: a b c --> (a 0 0) (b 0 0) (c 0 0)
            # see docstring above for format (!)
            # Treat empty 3 entry boxes as not having a unit cell
            if np.allclose(unitcell, [0.0, 0.0, 0.0]):
                wmsg = (
                    "Empty box [0., 0., 0.] found - treating as missing "
                    "unit cell. Dimensions set to `None`."
                )
                warnings.warn(wmsg)
                self.ts.dimensions = None
            else:
                self.ts.dimensions = np.r_[unitcell, [90.0, 90.0, 90.0]]
        elif len(unitcell) == 9:
            self.ts.dimensions = _gmx_to_dimensions(unitcell)
        else:  # raise an error for wrong format
            errmsg = "GRO unitcell has neither 3 nor 9 entries."
            raise ValueError(errmsg)

        if self.convert_units:
            self.convert_pos_from_native(self.ts._pos)  # in-place !
            if self.ts.dimensions is not None:
                self.convert_pos_from_native(
                    self.ts.dimensions[:3]
                )  # in-place!
            if self.ts.has_velocities:
                # converts nm/ps to A/ps units
                self.convert_velocities_from_native(self.ts._velocities)

    def Writer(self, filename, n_atoms=None, **kwargs):
        """Returns a CRDWriter for *filename*.

        Parameters
        ----------
        filename: str
            filename of the output GRO file

        Returns
        -------
        :class:`GROWriter`

        """
        if n_atoms is None:
            n_atoms = self.n_atoms
        return GROWriter(filename, n_atoms=n_atoms, **kwargs)


class GROWriter(base.WriterBase):
    """GRO Writer that conforms to the Trajectory API.

    Will attempt to write the following information from the topology:
     - atom name (defaults to 'X')
     - resnames (defaults to 'UNK')
     - resids (defaults to '1')


    .. note::
        The precision is hard coded to three decimal places.


    .. note::
        When dimensions are missing (i.e. set to `None`), a zero width
        unit cell box will be written (i.e. [0.0, 0.0, 0.0]).


    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based
    .. versionchanged:: 0.13.0
       Now strictly writes positions with 3dp precision.
       and velocities with 4dp.
       Removed the `convert_dimensions_to_unitcell` method,
       use `Timestep.triclinic_dimensions` instead.
       Now now writes velocities where possible.
    .. versionchanged:: 0.18.0
       Added `reindex` keyword argument to allow original atom
       ids to be kept.
    .. versionchanged:: 2.0.0
       Raises a warning when writing timestep with missing dimension
       information (i.e. set to ``None``).
    """

    format = "GRO"
    units = {"time": None, "length": "nm", "velocity": "nm/ps"}
    gro_coor_limits = {"min": -999.9995, "max": 9999.9995}

    #: format strings for the GRO file (all include newline); precision
    #: of 3 decimal places is hard-coded here.
    fmt = {
        "n_atoms": "{0:5d}\n",  # number of atoms
        # coordinates output format, see http://chembytes.wikidot.com/g-grofile
        "xyz": "{resid:>5d}{resname:<5.5s}{name:>5.5s}{index:>5d}{pos[0]:8.3f}{pos[1]:8.3f}{pos[2]:8.3f}\n",
        # unitcell
        "box_orthorhombic": "{box[0]:10.5f} {box[1]:9.5f} {box[2]:9.5f}\n",
        "box_triclinic": "{box[0]:10.5f} {box[4]:9.5f} {box[8]:9.5f} {box[1]:9.5f} {box[2]:9.5f} {box[3]:9.5f} {box[5]:9.5f} {box[6]:9.5f} {box[7]:9.5f}\n",
    }
    fmt["xyz_v"] = (
        fmt["xyz"][:-1] + "{vel[0]:8.4f}{vel[1]:8.4f}{vel[2]:8.4f}\n"
    )

    def __init__(self, filename, convert_units=True, n_atoms=None, **kwargs):
        """Set up a GROWriter with a precision of 3 decimal places.

        Parameters
        -----------
        filename : str
            output filename
        n_atoms : int (optional)
            number of atoms
        convert_units : bool (optional)
            units are converted to the MDAnalysis base format; [``True``]
        reindex : bool (optional)
            By default, all the atoms were reindexed to have a atom id starting
            from 1. [``True``] However, this behaviour can be turned off by
            specifying `reindex` ``=False``.

        Note
        ----
        To use the reindex keyword, user can follow the two examples given
        below.::

           u = mda.Universe()

        Usage 1::

           u.atoms.write('out.gro', reindex=False)

        Usage 2::

           with mda.Writer('out.gro', reindex=False) as w:
               w.write(u.atoms)

        """
        self.filename = util.filename(filename, ext="gro", keep=True)
        self.n_atoms = n_atoms
        self.reindex = kwargs.pop("reindex", True)

        self.convert_units = (
            convert_units  # convert length and time to base units
        )

    def write(self, obj):
        """Write selection at current trajectory frame to file.

        Parameters
        -----------
        obj : AtomGroup or Universe

        Note
        ----
        The GRO format only allows 5 digits for *resid* and *atom
        number*. If these numbers become larger than 99,999 then this
        routine will chop off the leading digits.


        .. versionchanged:: 0.7.6
           *resName* and *atomName* are truncated to a maximum of 5 characters
        .. versionchanged:: 0.16.0
           `frame` kwarg has been removed
        .. versionchanged:: 2.0.0
           Deprecated support for calling with Timestep has nwo been removed.
           Use AtomGroup or Universe as an input instead.
        """
        # write() method that complies with the Trajectory API

        try:

            # make sure to use atoms (Issue 46)
            ag = obj.atoms
            # can write from selection == Universe (Issue 49)

        except AttributeError:
            errmsg = "Input obj is neither an AtomGroup or Universe"
            raise TypeError(errmsg) from None

        try:
            velocities = ag.velocities
        except NoDataError:
            has_velocities = False
        else:
            has_velocities = True

        # Check for topology information
        missing_topology = []
        try:
            names = ag.names
        except (AttributeError, NoDataError):
            names = itertools.cycle(("X",))
            missing_topology.append("names")
        try:
            resnames = ag.resnames
        except (AttributeError, NoDataError):
            resnames = itertools.cycle(("UNK",))
            missing_topology.append("resnames")
        try:
            resids = ag.resids
        except (AttributeError, NoDataError):
            resids = itertools.cycle((1,))
            missing_topology.append("resids")

        if not self.reindex:
            try:
                atom_indices = ag.ids
            except (AttributeError, NoDataError):
                atom_indices = range(1, ag.n_atoms + 1)
                missing_topology.append("ids")
        else:
            atom_indices = range(1, ag.n_atoms + 1)
        if missing_topology:
            warnings.warn(
                "Supplied AtomGroup was missing the following attributes: "
                "{miss}. These will be written with default values. "
                "Alternatively these can be supplied as keyword arguments."
                "".format(miss=", ".join(missing_topology))
            )

        positions = ag.positions

        if self.convert_units:
            # Convert back to nm from Angstroms,
            # Not inplace because AtomGroup is not a copy
            positions = self.convert_pos_to_native(positions, inplace=False)
            if has_velocities:
                velocities = self.convert_velocities_to_native(
                    velocities, inplace=False
                )
        # check if any coordinates are illegal
        # (checks the coordinates in native nm!)
        if not self.has_valid_coordinates(self.gro_coor_limits, positions):
            raise ValueError(
                "GRO files must have coordinate values between "
                "{0:.3f} and {1:.3f} nm: No file was written."
                "".format(
                    self.gro_coor_limits["min"], self.gro_coor_limits["max"]
                )
            )

        with util.openany(self.filename, "wt") as output_gro:
            # Header
            output_gro.write("Written by MDAnalysis\n")
            output_gro.write(self.fmt["n_atoms"].format(ag.n_atoms))

            # Atom descriptions and coords
            # Dont use enumerate here,
            # all attributes could be infinite cycles!
            for atom_index, resid, resname, name in zip(
                range(ag.n_atoms), resids, resnames, names
            ):
                truncated_atom_index = util.ltruncate_int(
                    atom_indices[atom_index], 5
                )
                truncated_resid = util.ltruncate_int(resid, 5)
                if has_velocities:
                    output_gro.write(
                        self.fmt["xyz_v"].format(
                            resid=truncated_resid,
                            resname=resname,
                            index=truncated_atom_index,
                            name=name,
                            pos=positions[atom_index],
                            vel=velocities[atom_index],
                        )
                    )
                else:
                    output_gro.write(
                        self.fmt["xyz"].format(
                            resid=truncated_resid,
                            resname=resname,
                            index=truncated_atom_index,
                            name=name,
                            pos=positions[atom_index],
                        )
                    )

            # Footer: box dimensions
            if ag.dimensions is None or np.allclose(
                ag.dimensions[3:], [90.0, 90.0, 90.0]
            ):
                if ag.dimensions is None:
                    wmsg = (
                        "missing dimension - setting unit cell to zeroed "
                        "box [0., 0., 0.]"
                    )
                    warnings.warn(wmsg)
                    box = np.zeros(3)
                else:
                    box = self.convert_pos_to_native(
                        ag.dimensions[:3], inplace=False
                    )
                # orthorhombic cell, only lengths along axes needed in gro
                output_gro.write(self.fmt["box_orthorhombic"].format(box=box))
            else:
                try:  # for AtomGroup/Universe
                    tri_dims = obj.universe.coord.triclinic_dimensions
                except AttributeError:  # for Timestep
                    tri_dims = obj.triclinic_dimensions
                # full output
                box = self.convert_pos_to_native(
                    tri_dims.flatten(), inplace=False
                )
                output_gro.write(self.fmt["box_triclinic"].format(box=box))



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""GROMOS11 trajectory reader --- :mod:`MDAnalysis.coordinates.TRC`
====================================================================

Reads coordinates, timesteps and box-sizes from GROMOS11 TRC trajectories.

To load the trajectory into :class:`~MDAnalysis.core.universe.Universe`,
you need to provide topology information using a topology file such as
a PDB::

    import MDAnalysis as mda
    u = mda.Universe("topology.pdb", ["md_1.trc.gz","md_2.trc.gz"],
    continuous=True)

.. Note::
   The GROMOS trajectory format organizes its data in blocks. A block starts
   with a blockname in capital letters (example: POSITIONRED) and ends with
   a line containing only ''END''. Only the TITLE-block at the beginning of
   each file is mandatory, others blocks can be chosen depending on the task.

   The trajectory format as well as all permitted blocks and their data are
   documented in the GROMOS Manual Vol. 4, chapter 2 and 4.
   The manual can be downloaded here:
   https://gromos.net/gromos11_pdf_manuals/vol4.pdf

This reader is designed to read the blocks "TIMESTEP", "POSITIONRED" and
"GENBOX" from the trajectory which covers the standard trajectories of most
simulations . If a trajectory includes other blocks, a warning is served
and the blocks are ignored.

.. Note::
   MDAnalysis requires the blocks to be in the same order for each frame
   and ignores non-supported blocks.



Classes
-------

.. autoclass:: TRCReader
   :members:

"""

import pathlib
import errno
import warnings
import numpy as np

from . import base
from .timestep import Timestep
from ..lib import util
from ..lib.util import cached, store_init_arguments

import logging

logger = logging.getLogger("MDAnalysis.coordinates.GROMOS11")


class TRCReader(base.ReaderBase):
    """Coordinate reader for the GROMOS11 format"""

    format = "TRC"
    units = {"time": "ps", "length": "nm"}
    _Timestep = Timestep

    SUPPORTED_BLOCKS = ["TITLE", "TIMESTEP", "POSITIONRED", "GENBOX"]
    NOT_SUPPORTED_BLOCKS = [
        "POSITION",
        "REFPOSITION",
        "VELOCITY",
        "VELOCITYRED",
        "FREEFORCE",
        "FREEFORCERED",
        "CONSFORCE",
        "CONSFORCERED",
        "SHAKEFAILPOSITION",
        "SHAKEFAILPREVPOSITION",
        "LATTICESHIFTS",
        "COSDISPLACEMENTS",
        "STOCHINT",
        "NHCVARIABLES",
        "ROTTRANSREFPOS",
        "PERTDATA",
        "DISRESEXPAVE",
        "JVALUERESEXPAVE",
        "JVALUERESEPS",
        "JVALUEPERSCALE",
        "ORDERPARAMRESEXPAVE",
        "XRAYRESEXPAVE",
        "LEUSBIAS",
        "LEUSBIASBAS",
        "ENERGY03",
        "VOLUMEPRESSURE03",
        "FREEENERGYDERIVS03",
        "BFACTOR",
        "AEDSS",
    ]

    @store_init_arguments
    def __init__(self, filename, convert_units=True, **kwargs):
        super(TRCReader, self).__init__(filename, **kwargs)

        # GROMOS11 trajectories are usually either *.trc or *.trc.gz.
        # (trj suffix can come up when trajectory obtained from clustering)
        ext = pathlib.Path(self.filename).suffix
        if (ext[1:] == "trc") or (ext[1:] == "trj"):
            self.compressed = None
        else:
            self.compressed = ext[1:]

        self.trcfile = util.anyopen(self.filename)
        self.convert_units = convert_units

        # Read and calculate some information about the trajectory
        self.traj_properties = self._read_traj_properties()

        self._cache = {}
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)

        self._reopen()
        self.ts.dt = self.traj_properties["dt"]

        self._read_frame(0)

    @property
    @cached("n_atoms")
    def n_atoms(self):
        """The number of atoms in one frame."""
        return self.traj_properties["n_atoms"]

    @property
    @cached("n_frames")
    def n_frames(self):
        """The number of frames in the trajectory."""
        return self.traj_properties["n_frames"]

    def _frame_to_ts(self, frameDat, ts):
        """Convert a frame to a :class:`TimeStep`"""
        ts.frame = self._frame
        ts.time = frameDat["time"]

        ts.data["time"] = frameDat["time"]
        ts.data["step"] = frameDat["step"]

        ts.dimensions = frameDat["dimensions"]
        ts.positions = frameDat["positions"]

        # Convert the units
        if self.convert_units:
            if ts.has_positions:
                self.convert_pos_from_native(ts.positions)
            if ts.dimensions is not None:
                self.convert_pos_from_native(ts.dimensions[:3])

        return ts

    def _read_traj_properties(self):
        """
        * Reads the number of atoms per frame (n_atoms)
        * Reads the number of frames (n_frames)
        * Reads the startposition of the timestep block
          for each frame (l_blockstart_offset)
        """

        traj_properties = {}

        #
        # Check which of the supported blocks comes first in the trajectory
        #
        first_block = None
        with util.anyopen(self.filename) as f:
            for line in iter(f.readline, ""):
                stripped_line = line.strip()
                if stripped_line in TRCReader.SUPPORTED_BLOCKS and (
                    stripped_line != "TITLE"
                ):
                    # Save the name of the first non-title block
                    # in the trajectory file
                    first_block = stripped_line
                    break

        #
        # Calculate meta-data of the trajectory
        #
        n_atoms = 0
        frame_counter = 0
        block_size = {}
        block_size["POSITIONRED"] = None
        inconsistent_size = False

        l_blockstart_offset = []
        l_timestep_timevalues = []

        with util.anyopen(self.filename) as f:
            for line in iter(f.readline, ""):
                stripped_line = line.strip()
                #
                # First block of frame
                #
                if first_block == stripped_line:
                    l_blockstart_offset.append(f.tell() - len(line))
                    frame_counter += 1

                #
                # Timestep-block
                #
                if "TIMESTEP" == stripped_line:
                    l_timestep_timevalues.append(
                        float(f.readline().split()[1])
                    )
                    while stripped_line != "END":
                        stripped_line = f.readline().strip()
                #
                # Coordinates-block
                #
                if "POSITIONRED" == stripped_line:
                    # Only count the number of atoms in the first block
                    # Also stores the size of the POSITIONRED block which
                    # should be consistent in most cases.
                    if n_atoms == 0:
                        start = f.tell() - len(line)
                        while stripped_line != "END":
                            line = f.readline()
                            stripped_line = line.strip()
                            if len(stripped_line.split()) == 3:
                                n_atoms += 1
                        end = f.tell() - len(line)
                        block_size["POSITIONRED"] = end - start

                    elif inconsistent_size:
                        while stripped_line != "END":
                            stripped_line = f.readline().strip()

                    else:
                        # While it is allowed for trailing whitespace to exist
                        # in the trajectories, none of the GROMOS trajectory
                        # writers do actually produce trailing whitespace.
                        # By default we try to assume a consistent size of the
                        # POSITIONRED block. We then try to skip reading subsequent
                        # occurences of this block.
                        # If we end up somewhere unexpected, we will throw
                        # a warning and fall back to iterating over the file.

                        # instead of looping over the file, looking for an END
                        # we can seek to where the end of the block should be.
                        current_pos = f.tell() - len(line)
                        f.seek(
                            f.tell() - len(line) + block_size["POSITIONRED"]
                        )

                        # Check if we are at the correct position
                        # If not, set inconsistent_size to true and seek back
                        # to where we were before
                        if f.readline().strip() != "END":
                            inconsistent_size = True
                            warnmsg = f"Inconsistent POSITIONRED block size in file {f.name}. Falling back to slow reader."
                            warnings.warn(warnmsg, UserWarning)
                            logger.warning(warnmsg)
                            f.seek(current_pos)

        if frame_counter == 0:
            errormsg = (
                "No supported blocks were found within the GROMOS trajectory!"
            )
            logger.error(errormsg)
            raise ValueError(errormsg)

        traj_properties["n_atoms"] = n_atoms
        traj_properties["n_frames"] = frame_counter
        traj_properties["l_blockstart_offset"] = l_blockstart_offset

        if len(l_timestep_timevalues) >= 2:
            traj_properties["dt"] = (
                l_timestep_timevalues[1] - l_timestep_timevalues[0]
            )
        else:
            traj_properties["dt"] = 0
            warnmsg = "The trajectory does not contain TIMESTEP blocks!"
            warnings.warn(warnmsg, UserWarning)
            logger.warning(warnmsg)

        return traj_properties

    def _read_GROMOS11_trajectory(self):
        frameDat = {}
        frameDat["step"] = int(self._frame)
        frameDat["time"] = float(0.0)
        frameDat["positions"] = np.zeros(
            (self.traj_properties["n_atoms"], 3), dtype=np.float64
        )
        frameDat["dimensions"] = None
        self.periodic = False

        # Read the trajectory
        f = self.trcfile
        for line in iter(f.readline, ""):
            stripped_line = line.strip()
            if "TIMESTEP" == stripped_line:
                tmp_step, tmp_time = f.readline().split()
                frameDat["step"] = int(tmp_step)
                frameDat["time"] = float(tmp_time)

            elif "POSITIONRED" == stripped_line:
                i = 0
                buffer = []
                while True:
                    coords_str = f.readline()
                    if "#" in coords_str:
                        continue
                    elif "END" in coords_str:
                        break
                    else:
                        buffer.append(coords_str)
                        i += 1

                data = np.fromstring(
                    "".join(buffer),
                    sep=" ",
                    dtype=np.float64,
                    count=self.n_atoms * 3,
                )

                frameDat["positions"] = data.reshape(-1, 3)

                if i != self.traj_properties["n_atoms"]:
                    errormsg = f"Found {i} atoms in step {self._frame}, but expected {self.traj_properties['n_atoms']}"
                    logger.error(errormsg)
                    raise ValueError(errormsg)

            elif "GENBOX" == stripped_line:
                ntb_setting = int(f.readline())
                if ntb_setting == 0:
                    frameDat["dimensions"] = None
                    self.periodic = False

                elif ntb_setting in [1, 2]:
                    tmp_a, tmp_b, tmp_c = map(float, f.readline().split())
                    tmp_alpha, tmp_beta, tmp_gamma = map(
                        float, f.readline().split()
                    )
                    frameDat["dimensions"] = [
                        tmp_a,
                        tmp_b,
                        tmp_c,
                        tmp_alpha,
                        tmp_beta,
                        tmp_gamma,
                    ]
                    self.periodic = True

                    # gb_line3
                    if (
                        sum(abs(float(v)) for v in f.readline().split())
                        > 1e-10
                    ):
                        errormsg = (
                            "This reader doesnt't support a shifted origin!"
                        )
                        logger.error(errormsg)
                        raise ValueError(errormsg)

                    # gb_line4
                    if (
                        sum(abs(float(v)) for v in f.readline().split())
                        > 1e-10
                    ):
                        errormsg = "This reader doesnt't support yawed, pitched or rolled boxes!"
                        logger.error(errormsg)
                        raise ValueError(errormsg)

                else:
                    errormsg = "This reader does't support truncated-octahedral periodic boundary conditions"
                    logger.error(errormsg)
                    raise ValueError(errormsg)
                break

            elif any(
                non_supp_bn in line
                for non_supp_bn in TRCReader.NOT_SUPPORTED_BLOCKS
            ):
                for non_supp_bn in TRCReader.NOT_SUPPORTED_BLOCKS:
                    if non_supp_bn == stripped_line:
                        warnmsg = non_supp_bn + " block is not supported!"
                        warnings.warn(warnmsg, UserWarning)
                        logger.warning(warnmsg)

        return frameDat

    def _read_frame(self, i):
        """read frame i"""
        self._frame = i - 1

        # Move position in file just (-2 byte) before the start of the block
        self.trcfile.seek(
            self.traj_properties["l_blockstart_offset"][i] - 2, 0
        )

        return self._read_next_timestep()

    def _read_next_timestep(self):
        self._frame += 1
        if self._frame >= self.n_frames:
            errormsg = "Trying to go over trajectory limit"
            logger.error(errormsg)
            raise IOError(errormsg)

        raw_framedata = self._read_GROMOS11_trajectory()
        self._frame_to_ts(raw_framedata, self.ts)

        return self.ts

    def _reopen(self):
        """Close and reopen the trajectory"""
        self.close()
        self.open_trajectory()

    def open_trajectory(self):
        if self.trcfile is not None:
            errormsg = errno.EALREADY, "TRC file already opened", self.filename
            logger.error(errormsg)
            raise IOError(errormsg)

        # Reload trajectory file
        self.trcfile = util.anyopen(self.filename)

        # Reset ts
        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)

        # Set _frame to -1, so next timestep is zero
        self._frame = -1

        return self.trcfile

    def close(self):
        """Close the trc trajectory file if it was open."""
        if self.trcfile is not None:
            self.trcfile.close()
            self.trcfile = None



from packaging.version import Version


def norm_version(version_str: str):
    """
    Normalize an input version string in the same way that setuptools' dist
    does.

    Parameters
    ----------
    version_str : str
      A version string to normalize.

    Returns
    -------
    str
      Normalised version string
    """
    return str(Version(version_str))


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--file", type=str, help="file with version to parse")
    args = parser.parse_args()

    with open(args.file) as filed:
        ver = filed.readlines()[0].strip("\n")

    print(norm_version(version_str=ver))
